#!/usr/bin/env python3
import matplotlib.pyplot as plt
import so3g
from so3g.proj import coords, quat
import sotodlib.coords.planets as planets
from sotodlib import core
from sotodlib.toast.ops import sim_sso
from sotodlib.core import metadata
from sotodlib.io.metadata import write_dataset, read_dataset
from sotodlib.tod_ops.filters import high_pass_sine2, low_pass_sine2, fourier_filter
import pickle
from astropy import units as u
from copy import deepcopy
import numpy as np
import scipy.signal
import time
import ephem
import os
from tqdm.auto import tqdm
from scipy.optimize import curve_fit, minimize
from datetime import datetime
import argparse as ap
import yaml
from sotodlib.site_pipeline import util
from dataclasses import dataclass
import matplotlib
matplotlib.use('agg')

opj = os.path.join

logger = util.init_logger(__name__)


def find_source(
        tod,
        sso_name,
        beamsize=30,
        height=0.1,
        distance=10000,
        npeaks=3):
    """Find the detectors that see the source

    Args
    ------
    height: minimum required amplitude of a peak
    max_amp: over this amplitude it can only be a spike
    distance: distance between adjacent peaks -- depends on scan speed
    npeaks: number of required peaks to be found of certain amplitude
    beamsize: cluster of peaks with lower width than this is considered a spike

    Return
    ------
    good_dets : array of indices of the detectors that see the source
    """
    az = tod.boresight.az
    el = tod.boresight.el
    ctime = tod.timestamps

    good_dets = []

    for det_idx in range(len(tod.dets.vals)):

        data = tod.signal[det_idx, :]
        data -= np.mean(data)
        # find npeaks over some amplitude and sort them
        r, _ = scipy.signal.find_peaks(data, height=height, distance=distance)
        sorted_r = np.argsort(data[r])[::-1]
        # to avoid high amplitude spikes
        if len(r) < npeaks or np.nanmax(data) > 100 * height:
            continue

        peak_idxs = r[sorted_r[:npeaks]]
        peak_idxs_sorted = np.sort(peak_idxs)

        try:
            argmax = np.argmax(data[peak_idxs_sorted])
        except BaseException:
            continue

        # Separate between peaks left and right from the highest peak
        left_d = [data[i] for i in peak_idxs_sorted[:argmax + 1]]
        right_d = [data[i] for i in peak_idxs_sorted[argmax:]]
        # Assert the monotony changes at the highest peak to make sure
        # the source center is included
        if len(left_d) >= 2 and len(right_d) >= 2:
            if np.all(
                left_d == np.sort(left_d)) and np.all(
                right_d == np.sort(right_d)[
                    ::-1]):
                # Turn to source-centred coordinates
                all_coords = get_xieta_src_centered(
                    ctime, az, el, data, sso_name=sso_name, threshold=height, distance=distance)
                [xi, eta], [xi_det, eta_det] = all_coords
                # Attempt an estimation of the width to determine if we
                # are seeing a point source crossing or random spikes
                max_peak = np.nanmax(data) / 2
                sorted_idx = np.argsort(np.abs(data - max_peak))[:1000]

                xi_min = np.nanmin(xi[sorted_idx])
                xi_max = np.nanmax(xi[sorted_idx])
                eta_min = np.nanmin(eta[sorted_idx])
                eta_max = np.nanmax(eta[sorted_idx])

                fwhm_xi = np.degrees(xi_max - xi_min) * 60
                fwhm_eta = np.degrees(eta_max - eta_min) * 60

                # If it is smaller than 3/4 * beam size is a spike
                # we can't do narrower than the diffraction limit
                # + some effect from the filtering
                if fwhm_xi < 3 / 4 * beamsize and fwhm_eta < 3 / 4 * beamsize:
                    continue

                xi_det_center = xi - xi_det
                eta_det_center = eta - eta_det
                radius_main = np.radians(2.5)
                radius = np.sqrt(xi_det_center ** 2 + eta_det_center ** 2)
                idx_out = np.where(radius > radius_main)[0]
                snr = np.nanmax(data) / np.nanstd(data[idx_out])

                # Something extra we can do but not necessary - targetted to Moon data
                # Get the first sorted 20 peaks and assert the difference in amplitude
                # between the first and last is big (otherwise we caught some periodical
                # spikes and were unlucky enough to agree with the previous criteria)
                # r, _ = scipy.signal.find_peaks(data, height=height/3, distance=10000)
                # sorted_r = np.argsort(data[r])[::-1]

                # if len(r)<npeaks:
                #     continue

                # peak_idxs = r[sorted_r[:20]]
                # ratio = data[peak_idxs[-1]]/data[peak_idxs[0]]

                # As a subsequent sanity check look at the SNR
                if snr > 25:
                    good_dets.append(det_idx)

    return good_dets


def get_xieta_src_centered(
        ctime,
        az,
        el,
        data,
        sso_name,
        threshold,
        distance,
        prior_pos=None):
    """Create a planet centered coordinate system for single detector data

    Args
    ------
    ctime, az, el: unix time and boresight position coordinates
    data: single detector data
    sso_name: celestial source to insert to pyephem
    threshold: minimal peak amplitude required

    Return
    ------
    xi_total, eta_total : source's centre projected on the focal plane coords
    xi0, eta0 : source's centre projected on the detector coords
    """

    csl = so3g.proj.CelestialSightLine.az_el(ctime, az, el, weather="typical")
    q_bore = csl.Q

    # Signal strength criterion - a chance to flag detectors
    peaks, _ = scipy.signal.find_peaks(
        data, height=threshold, distance=distance)
    if not list(peaks):
        return

    src_idx = np.where(data == np.max(data[peaks]))[0][0]

    # planet position
    d1_unix = ctime[src_idx]
    planet = planets.SlowSource.for_named_source(sso_name, d1_unix * 1.)
    ra0, dec0 = planet.pos(d1_unix)
    q_obj = so3g.proj.quat.rotation_lonlat(ra0, dec0)

    # find max point in term of the source's coord system if no prior
    # positions arrer given
    if prior_pos is not None:
        xi0, eta0 = prior_pos
    else:
        q_det = coords.quat.quat(*np.array(~q_bore * q_obj)[src_idx])
        xi0, eta0, gamma0 = quat.decompose_xieta(q_det)

    q_total = ~q_bore * q_obj
    xi, eta, gamma = quat.decompose_xieta(q_total)

    return np.array([xi, eta]), np.array([xi0, eta0])


def define_fit_params(config_file_path, band, tele, tube, sso_name):
    """
    Define initial position and allowed offsets for detector centroids,
    beam parameters, and time constant as read from configuration file
    located at 'config_file_path'.

    Args
    ------
    tele, tube : telescope name and tube name
    sso_name: source name

    Return
    ------
    initial_guess, bounds: arguments for the least square fits
                                    and beam size in radians.
    tau: time constant value set from the configuration file.
    """
    with open(config_file_path, "r") as stream:
        config_file = yaml.safe_load(stream)
    tube_idx = config_file['telescopes'][tele]['tube'].index(tube)

    # Pointing - initial guess at zero
    offset_bound = config_file['telescopes'][tele]['pointing_error'][tube_idx]
    # Beam
    beamsize = np.radians(
        config_file['telescopes'][tele]['beamsize'][tube_idx] / 60)
    fwhm_bound_frac = config_file['telescopes'][tele]['beam_error'][tube_idx]
    # Amplitude
    amp = config_file['telescopes'][tele]['detector_response'][sso_name][tube_idx]
    amp_error = config_file['telescopes'][tele]['amp_error'][tube_idx]
    # Beam angle
    theta_min, theta_init, theta_max = np.radians(
        np.array(config_file['telescopes'][tele]['theta']))
    # Time constant
    tau = config_file['telescopes'][tele]['tau']

    initial_guess = [amp, 0, 0, beamsize, beamsize, theta_init]
    errors = [amp_error, offset_bound, offset_bound, fwhm_bound_frac,
              fwhm_bound_frac, theta_min, theta_max]

    return initial_guess, errors, tau


def set_bounds(initial_guess, errors):
    """Synthesize the bounds from the initial guess
       and allowed errors"""
    amp, xi0, eta0, fwhm_xi, fwhm_eta, theta_init = initial_guess
    amp_error, offset_bound, offset_bound, fwhm_bound_frac, fwhm_bound_frac, theta_min, theta_max = errors

    fwhm_min_xi = fwhm_xi - (fwhm_bound_frac * fwhm_xi)
    fwhm_max_xi = fwhm_xi + (fwhm_bound_frac * fwhm_xi)
    fwhm_min_eta = fwhm_eta - (fwhm_bound_frac * fwhm_eta)
    fwhm_max_eta = fwhm_eta + (fwhm_bound_frac * fwhm_eta)

    bounds_min = (amp - (amp_error * amp), -offset_bound, -
                  offset_bound, fwhm_min_xi, fwhm_min_eta, theta_min)
    bounds_max = (amp + (amp_error * amp), offset_bound,
                  offset_bound, fwhm_max_xi, fwhm_max_eta, theta_max)
    bounds = np.array((bounds_min, bounds_max,))

    return bounds


def gaussian2d(xi, eta, a, xi0, eta0, fwhm_xi, fwhm_eta, phi):
    """Simulate a time stream with an Gaussian beam model
    Args
    ------
    xi, eta: cordinates in the detector's system
    a: float
        amplitude of the Gaussian beam model
    xi0, eta0: float, float
        center position of the Gaussian beam model
    fwhm_xi, fwhm_eta, phi: float, float, float
        fwhm along the xi, eta axis (rotated)
        and the rotation angle (in radians)

    Ouput:
    ------
    sim_data: 1d array of float
        Time stream at sampling points given by xieta
    """
    xi_rot = xi * np.cos(phi) - eta * np.sin(phi)
    eta_rot = xi * np.sin(phi) + eta * np.cos(phi)
    factor = 2 * np.sqrt(2 * np.log(2))
    xi_coef = -0.5 * (xi_rot - xi0) ** 2 / (fwhm_xi / factor) ** 2
    eta_coef = -0.5 * (eta_rot - eta0) ** 2 / (fwhm_eta / factor) ** 2
    sim_data = a * np.exp(xi_coef + eta_coef)
    return sim_data


def tconst_convolve(ctime, data, tau):
    """Convolve the time stream with a given time constant
    Args
    ------
    ctime: 1d array of float
        containting the ctime of each data point
    data: 1d array of float
        time stream data for a given detector
    tau: float
        time constant to be added to the time stream (in seconds)

    Return
    ------
    data_tau: 1d array of float
        time-constant filtered time stream
    """
    k = int(np.ceil(np.log(len(data)) / np.log(2)))
    n = 2 ** k
    delta_time = np.mean(np.diff(ctime))
    freqs = np.fft.fftfreq(n, delta_time)
    data_pad = np.zeros(n)
    data_pad[: len(data)] = data
    fft_data = np.fft.fft(data_pad)
    fft_data *= 1.0 / (1.0 + 2.0j * np.pi * tau * freqs)
    data_tau = np.fft.ifft(fft_data).real[: len(data)]
    return data_tau


def tod_sim(xietat, a, xi0, eta0, fwhm_xi, fwhm_eta, phi, idx_main, tau):
    """ Simulate the TOD of a single detector
    Args
    ----
    xietat: stacked array of xi, eta, ctime
    a, xi0, eta0, fwhm_xi, fwhm_eta, phi: all floats
        parameters for the beam model, as elaborated in the gaussian2d function
    idx_main: range of integers serving as data indices to select a region around
              the peak of radius = 2*FWHM (reduce data volume - speed the fitting).
    tau: float
        time constant
    """
    xi, eta, ctime = xietat
    data_sim = gaussian2d(
        xi, eta, a, xi0, eta0, fwhm_xi, fwhm_eta, phi
    )

    if tau == 0.0:
        return data_sim[idx_main]
    data_sim_tau = tconst_convolve(ctime, data_sim, tau)
    return data_sim_tau[idx_main]


def fit_params(
        readout_id,
        data,
        ctime,
        az,
        el,
        band,
        sso_name,
        threshold_src,
        distance,
        init_params,
        fit_pointing=False,
        fit_beam=False,
        prior_pointing=None,
        representative_dets='no_detectors',
        **kwargs):
    """Function that fits individual time-streams and returns the readout id,
       peak amplitude, detector centroids, beam size in both directions and
       corresponding angle along with the estimated signal-to-noise ratio as
       an array

    Args
    ----
    readout_id: str
        Detector id string.
    data: array of float
        Vector of time-ordered data for this detector.
    ctime: array of float
        timestamps
    az: array of float
        azimuth of boresight
    el: array of float
        elevation of boresight
    sso_name: string
        Name of the celestial source
    init_params: arrays of floats; the output of 'define_fit_params'
    representative_dets: readout ids of detectors whose raw and fitted
                         timestreams should be plotted to evaluate
                         the fitting quality.
    """
    failed_params = np.array([readout_id, *np.full(7, np.nan)])

    data -= np.mean(data)

    if prior_pointing is not None and readout_id in prior_pointing.keys():
        prior_pos = [
            prior_pointing[readout_id]['xi0'],
            prior_pointing[readout_id]['eta0']]
    else:
        prior_pos = None

    coord_transforms = get_xieta_src_centered(
        ctime, az, el, data, sso_name, threshold_src, distance, prior_pos=prior_pos)

    if coord_transforms is None:
        return failed_params

    [xi_det_center, eta_det_center], [xi0, eta0] = coord_transforms
    # centre around detector - no angle information needed
    xi_det_center -= xi0
    eta_det_center -= eta0

    p0, errors, tau = init_params

    if sso_name == 'Moon':
        max_peak = np.nanmax(data) / 2
        sorted_idx = np.argsort(np.abs(data - max_peak))[:1000]

        xi_min = np.nanmin(xi_det_center[sorted_idx])
        xi_max = np.nanmax(xi_det_center[sorted_idx])

        eta_min = np.nanmin(eta_det_center[sorted_idx])
        eta_max = np.nanmax(eta_det_center[sorted_idx])

        fwhm_xi = xi_max - xi_min
        fwhm_eta = eta_max - eta_min

        p0[3:5] = fwhm_xi, fwhm_eta

        # Set the pointing error (optionally) as the difference
        # between two adjacent peaks / detector hits on the source
        # peak_idxs, _ = scipy.signal.find_peaks(data, distance=100)
        # peak_idxs_sorted = np.argsort(data[peak_idxs])[::-1][:6]
        # offset_bound_xi = np.abs(np.mean(np.diff(xi_det_center[peak_idxs_sorted])))
        # offset_bound_eta = np.abs(np.mean(np.diff(eta_det_center[peak_idxs_sorted])))

    # Trim the data around the source
    beamsize = np.nanmax([p0[3], p0[4]])
    radius_main = 2 * beamsize
    radius_cut = 2.5 * radius_main

    radius = np.sqrt(xi_det_center ** 2 + eta_det_center ** 2)
    # take a small piece of the data around the peak
    idx_main_in = np.where(radius < radius_main)[0]
    idx_main_out = np.where(radius > radius_main)[0]
    xi_main = xi_det_center[np.array(idx_main_in)]
    eta_main = eta_det_center[idx_main_in]
    data_main = data[idx_main_in]

    xi_eta_ctime = np.vstack((xi_det_center, eta_det_center, ctime))

    # append amplitude of max peak - safer for now
    # until we have insanely strict criteria on the source's amplitude/
    # efficiency
    p0[0] = np.nanmax(data_main)
    bounds = set_bounds(p0, errors)

    if fit_pointing is True:
        #  Fit the pointing
        f = lambda xyt, *pointing: tod_sim(
            xyt, pointing[0], pointing[1], pointing[2], p0[3], p0[4], p0[5], idx_main_in, tau)
        try:
            popt_pointing, _ = curve_fit(
                f, xi_eta_ctime, data_main, p0=p0[:3], bounds=bounds[:, :3])
            # logger.warning(f"fitted pointing {} {}".format(p0,bounds))
        except BaseException:
            return failed_params

        # Re-centering the data after fitting for the pointing
        p0[0] = popt_pointing[0]
        xi_det_center += popt_pointing[1]
        eta_det_center += popt_pointing[2]
        xi_eta_ctime = np.vstack((xi_det_center, eta_det_center, ctime))

        radius = np.sqrt(xi_det_center ** 2 + eta_det_center ** 2)
        idx_main_in = np.where(radius < radius_main)[0]
        idx_main_out = np.where(radius > radius_main)[0]
        xi_main = xi_det_center[idx_main_in]
        eta_main = eta_det_center[idx_main_in]
        data_main = data[idx_main_in]

        amp = popt_pointing[0]
        # The detector centroids are the initial positions we determined from the
        # peak amplitude plus the fitted offsets
        xi0_fitted = xi0 + popt_pointing[1]
        eta0_fitted = eta0 + popt_pointing[2]

    if fit_beam is True:
        # Fit the beam parameters
        f = lambda xyt, * \
            beam: tod_sim(xyt, p0[0], p0[1], p0[2], beam[0], beam[1], beam[2],
                          idx_main_in, tau)
        try:
            popt_beam, _ = curve_fit(
                f, xi_eta_ctime, data_main, p0=p0[3:], bounds=bounds[:, 3:])
            # logger.warning(f"fitted beam {} {}".format(p0,bounds))
        except BaseException:
            return failed_params

        fwhm_xi, fwhm_eta, phi = popt_beam

    # Estimate the noise as the sigma of the data outside a region of
    # radius>radius_cut
    noise = np.nanstd(data[np.where(radius > radius_cut)[0]])
    snr = amp / noise

    # If pointing / beam were not fitted return the initial guess
    # for this parameters
    if fit_pointing is False:
        amp, xi0_fitted, eta0_fitted = p0[0], xi0, eta0

    if fit_beam is False:
        fwhm_xi, fwhm_eta, phi = p0[3:6]

    # Plot the raw vs fitted tod for some representative detectors
    if readout_id in representative_dets:
        xi_det_center += popt_pointing[1]
        eta_det_center += popt_pointing[2]
        radius = np.sqrt(xi_det_center ** 2 + eta_det_center ** 2)
        idx_main_in = np.where(radius < radius_main)[0]
        xi_main = xi_det_center[idx_main_in]
        eta_main = eta_det_center[idx_main_in]
        data_main = data[idx_main_in]

        data_sim_main = gaussian2d(
            xi_main,
            eta_main,
            amp,
            0,
            0,
            fwhm_xi,
            fwhm_eta,
            phi)

        plot_tods(
            readout_id,
            xi_main,
            eta_main,
            data_main,
            data_sim_main,
            **kwargs)

    return np.array([readout_id, amp, xi0_fitted,
                    eta0_fitted, fwhm_xi, fwhm_eta, phi, snr])


def get_hw_positions(tele, band, tube=None):
    """ Get the hardware xi, eta, gamma positions and detector names
    that belong to a specific tube."""

    from sotodlib.sim_hardware import sim_nominal, sim_detectors_toast

    # Tube can be defined here for the SAT but should be provided for LAT
    tube_sat = {
        'f030': 'SAT4',
        'f040': 'SAT4',
        'f090': 'SAT1',
        'f150': 'SAT1',
        'f230': 'SAT3',
        'f290': 'SAT3',
    }

    if tube is None:
        tube = tube_sat['f' + str(band).zfill(3)]

    hw = sim_nominal()
    sim_detectors_toast(hw, tube)

    qdr, det_names_hw = [], []
    for names in hw.data['detectors'].keys():
        if band in names:
            det_names_hw.append(names)
            qdr.append([hw.data['detectors'][names]['quat'][3]] +
                       list(hw.data['detectors'][names]['quat'][:3]))

    quat_det = so3g.proj.quat.G3VectorQuat(np.array(qdr))
    xi_hw, eta_hw, gamma_hw = so3g.proj.quat.decompose_xieta(quat_det)

    return xi_hw, eta_hw, gamma_hw, np.array(det_names_hw)


def plot_planet_footprints(tod, sso_name, tele, tube, band, hw_pos, **kwargs):
    """ Plot the source's footprint on the focal plane """

    fig = plt.figure(dpi=300)
    ax = fig.add_subplot(111)

    xi_hw, eta_hw, _, dets_hw = hw_pos
    ax.plot(np.degrees(xi_hw), np.degrees(eta_hw), '.')

    csl = so3g.proj.CelestialSightLine.az_el(
        tod.timestamps, tod.boresight.az, tod.boresight.el, weather="typical")
    q_bore = csl.Q
    planet = planets.SlowSource.for_named_source(
        sso_name, tod.timestamps[0] * 1.)
    ra0, dec0 = planet.pos(tod.timestamps[0])
    planet_q = so3g.proj.quat.rotation_lonlat(ra0, dec0)
    q_total = ~q_bore * planet_q
    xi_total, eta_total, _ = quat.decompose_xieta(q_total)

    ax.plot(np.degrees(xi_total), np.degrees(
        eta_total), color='gray', alpha=.7)
    ax.set_xlabel('xi [deg]')
    ax.set_ylabel('eta [deg]')

    plt.savefig(opj(kwargs['img_dir'], 'source_footprint_' +
                kwargs['obs_id'] + '.png'), bbox_inches='tight')
    plt.close()


def make_fpu_plots(df, hw_pos,
                   input_param=None, fitted_par='beam', **kwargs):
    """ Make focal plane plots where the detectors are color-coded according
        to their (fitted pointing bias or) fitted beam size bias with respect
        to the input"""

    xi_hw, eta_hw, _, dets_hw = hw_pos
    df_det_idxs = [np.where(dets_hw == df['dets:readout_id'][i])[0][0]
                   for i in range(len(df['dets:readout_id']))]
    delta_xis = np.full(len(xi_hw), np.nan)
    delta_etas = np.full(len(eta_hw), np.nan)

    # if fitted_par == 'pointing':
    #     par1, par2 = 'xi0', 'eta0'
    #     x_hw_fitted, y_hw_fitted = xi_hw[df_det_idxs], eta_hw[df_det_idxs]

    # elif fitted_par == 'beam':
    par1, par2 = 'fwhm_xi', 'fwhm_eta'
    beamsize = input_param
    x_hw_fitted, y_hw_fitted = beamsize, beamsize

    x_df_fitted = df[par1].to_numpy().astype(float)
    y_df_fitted = df[par2].to_numpy().astype(float)

    delta_xis[df_det_idxs] = x_df_fitted - x_hw_fitted
    delta_etas[df_det_idxs] = y_df_fitted - y_hw_fitted

    fig, (ax1, ax2) = plt.subplots(1, 2, dpi=300)
    # We can make plotting options user specified
    v1_lim = np.max(np.abs(np.array(
        [np.nanmin(np.degrees(delta_xis)), np.nanmax(np.degrees(delta_xis))])))
    v2_lim = np.max(np.abs(np.array(
        [np.nanmin(np.degrees(delta_etas)), np.nanmax(np.degrees(delta_etas))])))

    im1 = ax1.scatter(np.degrees(xi_hw), np.degrees(eta_hw), c=np.degrees(
        delta_xis), vmin=-v1_lim, vmax=v1_lim, cmap='seismic')
    cbar = plt.colorbar(im1)
    cbar.set_label('Bias in ' + par1 + '[degrees]',
                   labelpad=25, rotation=270, size=13)
    ax1.set_xlabel('ξ[degrees]', size=13)
    ax1.set_ylabel('η[degrees]', size=13)

    im2 = ax2.scatter(np.degrees(xi_hw), np.degrees(eta_hw), c=np.degrees(
        delta_etas), vmin=-v2_lim, vmax=v2_lim, cmap='seismic')
    cbar = plt.colorbar(im2)
    cbar.set_label('Bias in ' + par2 + '[degrees]',
                   labelpad=25, rotation=270, size=13)
    ax2.set_xlabel('ξ[degrees]', size=13)
    ax2.set_ylabel('η[degrees]', size=13)

    fig.subplots_adjust(wspace=0.5)
    plt.savefig(opj(kwargs['img_dir'], fitted_par + '_bias_' +
                kwargs['obs_id'] + '.png'), bbox_inches='tight')
    plt.close()


def plot_tods(readout_id, xi_main, eta_main, data_main, data_sim_main,
              **kwargs):
    """
    Plot raw vs fitted timestream in as data=f(xi,eta) and
    show the detector's position on the fpu provided a set
    of detector names"""

    fig, (ax1, ax2) = plt.subplots(
        1, 2, figsize=(10, 4), dpi=300)
    ax1.plot(xi_main, data_main, lw=2, label='Raw data')
    ax1.plot(xi_main, data_sim_main, '--', lw=1)
    ax1.set_xlabel(r"$\xi$[deg]")
    ax1.set_ylabel('T(K)')
    ax1.legend(loc='upper left', frameon=False)
    ax2.plot(eta_main, data_main, lw=2)
    ax2.plot(eta_main, data_sim_main, '--', lw=1, label='Fitted data')
    ax2.set_xlabel(r"$\eta$[deg]")
    ax2.set_ylabel('T(K)')
    ax2.legend(loc='upper left', frameon=False)
    # ax3.scatter(xi_hw, eta_hw, c='gray')
    # ax3.scatter(xi_hw[det_idx], eta_hw[det_idx], c='r')
    # ax3.set_xlabel(r"$\xi$[rad]")
    # ax3.set_ylabel(r"$\eta$[rad]")
    fig.subplots_adjust(wspace=0.31)
    plt.savefig(opj(kwargs['img_dir'], 'tod_det_' +
                    readout_id +
                    '_obsid_' +
                    kwargs['obs_id'] +
                    '.png'), bbox_inches='tight')
    plt.close()


@dataclass
class QuickFitResult:
    """
    results for pointing_quickfit function

    Args
    ----
    xi: float
        fitted detector xi [rad] relative to boresight center
    eta: float
        fitted detector eta [rad] relative to boresight center
    amp: float
        Amplitude used in the fit. This will be in the same units as
        ``am.signal``, which is generally detector phase.
    fwhm: float
        gaussian FHWM [rad] used in the fit.
    fit_aman: AxisManager
        AxisManager containing xi and eta of source relative to boresight center,
        fitted signal, and model.
    """
    xi: float
    eta: float
    amp: float
    fwhm: float
    fit_aman: core.AxisManager

def pointing_quickfit(
        am_full, chan_idx, downsample_factor=20, bandpass_range=(0.01, 1.8),
        fwhm=np.deg2rad(0.5), max_rad=None
    ):
    """
    Fits pointing for a bright source such as the moon for a single detector.

    Args
    -----
    am_full: AxisManager
        AxisManager containing unfiltered detector signal. This won't be modified.
    chan_idx: int
        Channel index in the axis manager to fit
    downsample_factor: int
        If set, the signal and model will be downsampled by this factor when fitting.
    bandpass_range: tuple[float, float]
        Low and high cutoff frequencies [Hz, Hz] for bandpass. If either are None, will not apply
        the corresponding highpass/lowpass.
    fwhm: float
        Full width half max [radians] of the beam to use for the model. This should be roughly the beam
        size of the telescope.
    max_rad: float
        Only data where the angular distance [radians] between the src and the estimated det location
        is less than ``max_rad`` will be used in the fit. If None, will use 5 *
        fwhm.

    Returns
    --------
        result: QuickFitResult
            Result object containing fitted xi and eta, along with amp and fwhm
            used in the fit.
    """

    if max_rad is None:
        max_rad = 5 * fwhm

    am = am_full.restrict('dets', [am_full.dets.vals[chan_idx]], in_place=False)

    def filter_tod(am, signal_name='signal'):
        filt_kw = dict(
            detrend='linear', resize='zero_pad', axis_name='samps',
            signal_name=signal_name, time_name='timestamps'
        )
        if bandpass_range[0] is not None:
            highpass = high_pass_sine2(cutoff=bandpass_range[0])
            am[signal_name] = fourier_filter(am, highpass, **filt_kw)
        if bandpass_range[1] is not None:
            lowpass = low_pass_sine2(cutoff=bandpass_range[1])
            am[signal_name] = fourier_filter(am, lowpass, **filt_kw)
        return am

    filter_tod(am)
    sl = slice(None, None, downsample_factor)

    ts = am.timestamps[sl]
    az = am.boresight.az[sl]
    el = am.boresight.el[sl]
    sig = am.signal[0, sl]

    xieta_src_rel_fp, xieta_det_rel_fp_est = get_xieta_src_centered(ts, az, el, sig, 'moon', 5)
    xieta_src_rel_det = xieta_src_rel_fp - xieta_det_rel_fp_est[:, None]

    m = np.sqrt(xieta_src_rel_det[0] ** 2 + xieta_src_rel_det[1] ** 2) < max_rad
    amp = np.ptp(sig[m]) * 3

    model_tod = gaussian2d(
        xieta_src_rel_det[0, m], xieta_src_rel_det[1, m], amp, 0, 0, fwhm, fwhm, 0
    )
    fit_am = core.AxisManager().wrap('timestamps', ts[m], [(0, core.IndexAxis('samps'))])
    fit_am.wrap('signal', np.array([sig[m]]), [(0, deepcopy(am.dets)), (1, 'samps')])
    fit_am.wrap('model', np.array([model_tod]), [(0, 'dets'), (1, 'samps')])

    def fit_func(xi0, eta0):
        fit_am.model[0] = gaussian2d(
            xieta_src_rel_det[0, m], xieta_src_rel_det[1, m], amp, xi0, eta0, fwhm, fwhm, 0
        )
        filter_tod(fit_am, signal_name='model')
        return np.sum((fit_am.signal[0] - fit_am.model[0])**2)

    res = minimize(lambda x: fit_func(*x), [0, 0])

    xi_det_rel_fp = xieta_det_rel_fp_est[0] + res.x[0]
    eta_det_rel_fp = xieta_det_rel_fp_est[1] + res.x[1]

    fit_am.model[0] = gaussian2d(
        xieta_src_rel_det[0, m], xieta_src_rel_det[1, m], amp, res.x[0], res.x[1], fwhm, fwhm, 0
    )
    fit_am.wrap('xi', xieta_src_rel_fp[0, m])
    fit_am.wrap('eta', xieta_src_rel_fp[1, m])

    result = QuickFitResult(
        xi=xi_det_rel_fp,
        eta=eta_det_rel_fp,
        amp=amp,
        fwhm=fwhm,
        fit_aman=fit_am
    )
    return result

def plot_quickfit_res(result: QuickFitResult, eta_scale=50):
    """
    Makes cool pseudo-3d source plot
    """
    fig, ax = plt.subplots()
    fit = result.fit_aman
    ax.plot(fit.xi, fit.signal[0] + fit.eta* eta_scale, label='Signal')
    ax.plot(fit.xi, fit.model[0] + fit.eta * eta_scale, label='Model', alpha=0.8)
    ax.legend()
    ax.set_xlabel("Xi (rad, relative to boresight)")
    ax.set_ylabel("Amplitude + scale * eta")
    return fig, ax


def quickfit_and_save_pointing(output_path, am, chan_idxs, show_pb=False, h5_address='fit_results'):
    """
    Run pointing_quickfit for a list of dets and save results to a ResultSet

    Args
    -----
    output_path: str
        Path to h5 file to save results
    am: AxisManager
        Axis manager with unfiltered TODs
    chan_idxs: list
        List of channel indexes of the axis manager to fit pointing
    show_pb: bool
        If True, enable tqdm progress bar
    h5_address: str
        Address in the h5 to save result set to.
    """
    dtype = [
        ('readout_id', '<U40'),
        ('xi', 'f8'),
        ('eta', 'f8'),
    ]
    res_arr = np.array([], dtype=dtype)
    rset= core.metadata.ResultSet.from_friend(res_arr)
    for i in tqdm(chan_idxs, disable=(not show_pb)):
        try:
            result = pointing_quickfit(am, i)
        except Exception:
            continue
        rset.append(
            {'readout_id': am.det_info.readout_id[i], 'xi': result.xi, 'eta': result.eta}
        )
        write_dataset(rset, output_path, h5_address, overwrite=True)
    return rset


def main(
    ctx_file,
    obs_id,
    sso_name,
    band,
    tele,
    tube,
    ufm,
    max_samps,
    scan_rate,
    config_file_path,
    outdir,
    highpass,
    cutoff_high,
    lowpass,
    cutoff_low,
    threshold_src,
    distance,
    do_abs_cal,
    fit_pointing,
    fit_beam,
    pointing_dict=None,
    representative_dets='no_detectors',
    test_mode=False,
    plot_results=False,
):
    """
    Initiate an MPI environment and fit a simulation in the time domain
    """
    # Two local imports, to avoid docs depenency.
    import pandas as pd

    kwargs = dict()
    kwargs['obs_id'] = obs_id

    img_dir = opj(outdir, 'plots')
    if not os.path.exists(img_dir):
        os.makedirs(img_dir)

    kwargs['img_dir'] = img_dir

    t1 = time.time()

    # Get the initial parameters
    init_params = define_fit_params(
        config_file_path, band, tele, tube, sso_name)

    # Load prior on pointing if provided
    if pointing_dict is not None:
        prior_pointing = pickle.load(open(pointing_dict, 'rb'))
    else:
        prior_pointing = None

    # Set context, ufm, chunk size
    ctx = core.Context(ctx_file)
    obs = ctx.obsdb.query()

    stream_id = f'ufm_{ufm}'
    meta = ctx.get_meta(obs_id)

    if 'stream_id' in meta.det_info:
        meta.restrict(
            'dets',
            meta.dets.vals[meta.det_info.stream_id == stream_id]
        )

    if 'n_samps' in meta.obs_info:
        n_samps = meta.obs_info.n_samples
    else:
        d1 = datetime.fromtimestamp(meta.obs_info.start_time)
        d2 = datetime.fromtimestamp(meta.obs_info.stop_time)
        # If we are looking at simulation we need info on the scan
        # rate to estimate total number of samples from metadata
        n_samps = (d2 - d1).seconds * scan_rate

    n_chunks = n_samps // max_samps

    for i in range(n_chunks - 1):
        start = max_samps * i
        stop = max_samps * (i + 1)
        if i == max_samps - 1:
            stop = n_samps
        tod = ctx.get_obs(meta, samples=[start, stop], no_signal=False)

        logger.warning(f'Loading {obs_id} of {max_samps} samples for {ufm}')

        ctime, az, el = tod.timestamps, tod.boresight.az, tod.boresight.el

        # Filter
        if highpass and cutoff_high is not None:
            tod.signal = fourier_filter(
                tod,
                filt_function=high_pass_sine2(
                    cutoff=cutoff_high),
                detrend='linear',
                resize='zero_pad',
                axis_name='samps',
                signal_name='signal',
                time_name='timestamps')

        if lowpass and cutoff_low is not None:
            tod.signal = fourier_filter(
                tod,
                filt_function=low_pass_sine2(
                    cutoff=cutoff_low),
                detrend='linear',
                resize='zero_pad',
                axis_name='samps',
                signal_name='signal',
                time_name='timestamps')

        tod.signal[tod.signal < 0] = 0

        # Choose one band and calibrate to pW
        if 'det_match' in tod and 'det_cal' in tod:
            tod.restrict('dets',
                         tod.dets.vals[tod.det_match.det_bandpass == band])
            tod.signal = np.multiply(tod.signal.T, tod.det_cal.phase_to_pW).T

        # Find detectors with source
        dets_w_src = find_source(
            tod,
            sso_name=sso_name,
            height=threshold_src,
            distance=distance)
        rd_ids = tod.dets.vals[dets_w_src]
        logger.warning(f'got good dets')

        if len(dets_w_src) == 0:
            continue

        # Initiate a dataframe
        full_df = pd.DataFrame(
            columns=[
                "dets:readout_id",
                "amp",
                "xi0",
                "eta0",
                "fwhm_xi",
                "fwhm_eta",
                "phi",
                "snr",
            ], index=rd_ids,)

        count = 0
        for _i, rd_idx in enumerate(dets_w_src):
            logger.warning(f'Starting to fit the detectors')
            rd_id = rd_ids[_i]
            params = fit_params(rd_id, tod.signal[rd_idx, :],
                                ctime, az, el, band,
                                sso_name, threshold_src, distance, init_params,
                                fit_pointing, fit_beam, prior_pointing,
                                representative_dets, **kwargs)
            snr = float(params[-1])
            logger.info(f'Solved {rd_idx:<5d} "{rd_id}" with S/N={snr:.2f}')
            full_df.loc[_i, :] = np.array(params)
            count += 1
            if test_mode and count >= 2:
                break

        new_dtypes = {
            "dets:readout_id": str,
            "amp": np.float64,
            "xi0": np.float64,
            "eta0": np.float64,
            "fwhm_xi": np.float64,
            "fwhm_eta": np.float64,
            "phi": np.float64,
            "snr": np.float64,
        }

        full_df = full_df.astype(new_dtypes)

        # Store dataframe
        full_df.to_hdf(opj(outdir, 'parameter_fits_' + obs_id + '.h5'),
                       key='full_df', mode='w')

        # calculating relative and absolute calibration
        amp = full_df.amp.values
        full_df["rel_cal"] = amp / np.mean(amp)

        if do_abs_cal:
            beam_file = "/mnt/so1/shared/site-pipeline/bcp/%s_%s_beam.h5" % (
                tele, band)
            sso_obj = sim_sso.SimSSO(beam_file=beam_file, sso_name=sso_name)
            freq_arr_GHz, temp_arr = sso_obj._get_sso_temperature(sso_name)
            dt_obs = datetime.fromtimestamp(np.average(tod.timestamps))
            sso_ephem = getattr(ephem, sso_name)()
            sso_ephem.compute(dt_obs.strftime('%Y-%m-%d'))
            ttemp = np.interp(float(band[1:]) * u.GHz, freq_arr_GHz, temp_arr)
            beam, _ = sso_obj._get_beam_map(
                None, sso_ephem.size * u.arcsec, ttemp)
            amp_ref = beam(0, 0)[0][0]
            full_df["abs_cal"] = amp / amp_ref

        # Assert tables dependency is properly satisfied
        out_folder = opj(outdir, tele)
        if not os.path.exists(out_folder):
            # Create a new directory because it does not exist
            os.makedirs(out_folder)
        result_arr = full_df.to_records(index=False)
        result_rs = metadata.ResultSet.from_friend(result_arr)
        print(opj(out_folder, 'cal_obs_%s.h5' % tele))
        write_dataset(result_rs, opj(out_folder, 'cal_obs_%s.h5' %
                      tele), obs_id, overwrite=True)

        t2 = time.time()
        print("Time to run fittings for %s is %.2f seconds." %
              (obs_id, t2 - t1))

        # Plot planet footprint, bias across focal plane
        if plot_results:
            hw_pos = get_hw_positions(tele, band)
            plot_planet_footprints(
                tod, sso_name, tele, tube, band, hw_pos, **kwargs)
            # make_fpu_plots(full_df, tele, band,
            #                input_param=None, fitted_par='pointing', **kwargs)
            make_fpu_plots(
                full_df,
                hw_pos=hw_pos,
                input_param=init_params[2],
                fitted_par='beam',
                **kwargs)


def get_parser(parser=None):
    if parser is None:
        parser = ap.ArgumentParser(
            formatter_class=ap.ArgumentDefaultsHelpFormatter)

    parser.add_argument(
        "--ctx_file",
        action="store",
        dest="ctx_file",
        help="The location of the context file.",
        type=str,
    )

    parser.add_argument(
        "--obs_id",
        action="store",
        dest="obs_id",
        required=True,
        help="Observation id in the context file.",
        type=str,
    )

    parser.add_argument(
        "--sso_name",
        action="store",
        dest="sso_name",
        required=True,
        help="Source name.",
        type=str,
    )

    parser.add_argument(
        "--band",
        action="store",
        dest="band",
        required=True,
        help="Frequency band.",
        type=str,
    )

    parser.add_argument(
        "--tele",
        action="store",
        dest="tele",
        required=True,
        help="Telescope name (SAT/LAT).",
        type=str,
    )

    parser.add_argument(
        "--tube",
        action="store",
        dest="tube",
        required=True,
        help="Telescope tube.",
        type=str,
    )

    parser.add_argument(
        "--ufm",
        action="store",
        dest="ufm",
        required=True,
        help="Telescope ufm.",
        type=str,
    )

    parser.add_argument(
        "--max_samps",
        action="store",
        dest="max_samps",
        required=True,
        help="Number of maximum samples to load",
        type=int,
    )

    parser.add_argument(
        "--scan_rate",
        action="store",
        dest="scan_rate",
        required=False,
        default=200,
        help="Scan rate, required for simulations",
        type=int,
    )

    parser.add_argument(
        "--config_file_path",
        action="store",
        dest="config_file_path",
        required=True,
        help="Location of configuration file that contains beam size \
              and fitting parameters.",
        type=str,
    )

    parser.add_argument(
        "--outdir",
        action="store",
        dest="outdir",
        help="The location for the .h5 output files to be stored.",
    )

    parser.add_argument(
        "--test-mode",
        action="store_true",
        help="Run analysis on a subset of detectors, to quickly check \
                for problems.")

    parser.add_argument(
        "--highpass",
        action="store_true",
        dest="highpass",
        help="If True, use highpass sine filter.",
    )

    parser.add_argument(
        "--cutoff_high",
        action="store",
        dest="cutoff_high",
        default=None,
        help="The cutoff frequency to be used in the filtering.",
        type=float,
    )

    parser.add_argument(
        "--lowpass",
        action="store_true",
        dest="lowpass",
        help="If True, use lowpass sine filter.",
    )

    parser.add_argument(
        "--cutoff_low",
        action="store",
        dest="cutoff_low",
        default=None,
        help="The cutoff frequency to be used in the filtering.",
        type=float,
    )

    parser.add_argument(
        "--threshold_src",
        action="store",
        dest="threshold_src",
        default=10,
        help="The max amplitude required for the peak finding",
        type=float,
    )

    parser.add_argument(
        "--distance",
        action="store",
        dest="distance",
        default=50000,
        help="The max amplitude required for the peak finding",
        type=float,
    )

    parser.add_argument(
        "--do_abs_cal",
        action="store_true",
        dest="do_abs_cal",
        default=False,
        help="Do absolute calibration fit.",
    )

    parser.add_argument(
        "--fit_pointing",
        action="store_true",
        dest="fit_pointing",
        default=False,
        help="Fit the pointing parameters.",
    )

    parser.add_argument(
        "--fit_beam",
        action="store_true",
        dest="fit_beam",
        default=False,
        help="Fit the beam parameters.",
    )

    parser.add_argument(
        "--pointing_dict",
        action="store",
        dest="pointing_dict",
        required=False,
        type=str,
        help="Path to a pickle file of a pointing dictionary.",
    )

    parser.add_argument(
        "--representative_dets",
        action="store",
        dest="representative_dets",
        default='no_detectors',
        nargs='+',
        help="Representative detectors across the focal plane whose \
              raw and fitted data should be plotted, given as a list of \
              readout ids.",
    )

    parser.add_argument(
        "--plot_results",
        action="store_true",
        dest="plot_results",
        default=False,
        help="Make plots of planet footprint and fitted results.",
    )

    return parser


if __name__ == "__main__":
    util.main_launcher(main, get_parser)



# Copyright (c) 2023-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import numpy as np
from astropy import units as u
import toast
import toast.ops
from toast.observation import default_values as defaults

from .. import ops as so_ops


def setup_pointing(operators):
    """Add commandline args and operators for pointing.

    This sets up operators different types of pointing matrix
    scenarios.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    # Detector quaternion pointing
    operators.append(
        toast.ops.PointingDetectorSimple(
            name="det_pointing_azel",
            boresight=defaults.boresight_azel,
            quats="quats_azel",
        )
    )
    operators.append(
        toast.ops.PointingDetectorSimple(
            name="det_pointing_radec",
            boresight=defaults.boresight_radec,
            quats="quats_radec",
        )
    )
    # Stokes weights
    operators.append(
        toast.ops.StokesWeights(name="weights_azel", weights="weights_azel", mode="IQU")
    )
    operators.append(toast.ops.StokesWeights(name="weights_radec", mode="IQU"))
    # Healpix pixelization
    operators.append(toast.ops.PixelsHealpix(name="pixels_healpix_radec"))
    operators.append(
        toast.ops.PixelsHealpix(name="pixels_healpix_radec_final", enabled=False)
    )
    # WCS pixelization in Az/El
    operators.append(
        toast.ops.PixelsWCS(
            name="pixels_wcs_azel",
            projection="CAR",
            resolution=(0.005 * u.degree, 0.005 * u.degree),
            submaps=1,
            auto_bounds=True,
            enabled=False,
        )
    )
    operators.append(toast.ops.PixelsWCS(name="pixels_wcs_azel_final", enabled=False))
    # WCS pixelization in RA/DEC
    operators.append(
        toast.ops.PixelsWCS(
            name="pixels_wcs_radec",
            projection="CAR",
            resolution=(0.005 * u.degree, 0.005 * u.degree),
            submaps=1,
            auto_bounds=True,
            enabled=False,
        )
    )
    operators.append(toast.ops.PixelsWCS(name="pixels_wcs_radec_final", enabled=False))


def select_pointing(job, otherargs, runargs, data):
    """Select pointing scheme and check for consistency

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    log = toast.utils.Logger.get()

    # Configured operators for this job
    job_ops = job.operators

    n_enabled_solve = np.sum(
        [
            job_ops.pixels_wcs_azel.enabled,
            job_ops.pixels_wcs_radec.enabled,
            job_ops.pixels_healpix_radec.enabled,
        ]
    )
    if n_enabled_solve != 1:
        msg = "Only one pixelization operator should be enabled for the solver."
        log.error(msg)
        raise RuntimeError(msg)

    n_enabled_final = np.sum(
        [
            job_ops.pixels_wcs_azel_final.enabled,
            job_ops.pixels_wcs_radec_final.enabled,
            job_ops.pixels_healpix_radec_final.enabled,
        ]
    )
    if n_enabled_final > 1:
        msg = "At most, one pixelization operator can be enabled "
        msg += "for the final binning."
        log.error(msg)
        raise RuntimeError(msg)

    # Configure Az/El and RA/DEC boresight and detector pointing and weights

    job_ops.det_pointing_azel.boresight = defaults.boresight_azel
    job_ops.det_pointing_radec.boresight = defaults.boresight_radec

    job_ops.pixels_wcs_azel.detector_pointing = job_ops.det_pointing_azel
    job_ops.pixels_wcs_radec.detector_pointing = job_ops.det_pointing_radec
    job_ops.pixels_healpix_radec.detector_pointing = job_ops.det_pointing_radec

    job_ops.pixels_wcs_azel_final.detector_pointing = job_ops.det_pointing_azel
    job_ops.pixels_wcs_radec_final.detector_pointing = job_ops.det_pointing_radec
    job_ops.pixels_healpix_radec_final.detector_pointing = job_ops.det_pointing_radec

    job_ops.weights_azel.detector_pointing = job_ops.det_pointing_azel
    job_ops.weights_radec.detector_pointing = job_ops.det_pointing_radec

    if len(data.obs) > 0 and defaults.hwp_angle in data.obs[0].shared:
        job_ops.weights_azel.hwp_angle = defaults.hwp_angle
        job_ops.weights_radec.hwp_angle = defaults.hwp_angle

    # Select Pixelization and weights for solve and final binning

    if job_ops.pixels_wcs_azel.enabled:
        job.pixels_solve = job_ops.pixels_wcs_azel
        job.weights_solve = job_ops.weights_azel
    elif job_ops.pixels_wcs_radec.enabled:
        job.pixels_solve = job_ops.pixels_wcs_radec
        job.weights_solve = job_ops.weights_radec
    else:
        job.pixels_solve = job_ops.pixels_healpix_radec
        job.weights_solve = job_ops.weights_radec
    job.weights_final = job.weights_solve

    if n_enabled_final == 0:
        # Use same as solve
        job.pixels_final = job.pixels_solve
    else:
        if job_ops.pixels_wcs_azel_final.enabled:
            job.pixels_final = job_ops.pixels_wcs_azel_final
        elif job_ops.pixels_wcs_radec_final.enabled:
            job.pixels_final = job_ops.pixels_wcs_radec_final
        else:
            job.pixels_final = job_ops.pixels_healpix_radec_final

    # Set up pointing matrices for binning operators, if they are
    # being used in this workflow.

    if hasattr(job_ops, "binner"):
        job_ops.binner.pixel_pointing = job.pixels_solve
        job_ops.binner.stokes_weights = job.weights_solve
        job_ops.binner.full_pointing = otherargs.full_pointing
    if hasattr(job_ops, "binner_final"):
        job_ops.binner_final.pixel_pointing = job.pixels_final
        job_ops.binner_final.stokes_weights = job.weights_final
        job_ops.binner_final.full_pointing = otherargs.full_pointing
        # If we are not using a different binner for our final binning, use the
        # same one as the solve.
        if not job_ops.binner_final.enabled:
            job_ops.binner_final = job_ops.binner

    # See if there is separate pointing for simulations
    if hasattr(job_ops, "det_pointing_azel_sim"):
        if not job_ops.det_pointing_azel_sim.enabled:
            job_ops.det_pointing_azel_sim = job_ops.det_pointing_azel
    if hasattr(job_ops, "det_pointing_radec_sim"):
        if not job_ops.det_pointing_radec_sim.enabled:
            job_ops.det_pointing_radec_sim = job_ops.det_pointing_radec



# Copyright (c) 2024-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import numpy as np

from toast.fft import convolve
from toast.utils import Logger, rate_from_times
from toast.traits import trait_docs, Int, Unicode
from toast.timing import function_timer
from toast.observation import default_values as defaults
from toast.ops import Operator

from sotodlib.tod_ops import filters


@trait_docs
class ReadoutFilter(Operator):
    """Apply a readout filter to the signal"""

    # Class traits

    API = Int(0, help="Internal interface version for this operator")

    times = Unicode(defaults.times, help="Observation shared key for timestamps")

    det_data = Unicode(
        defaults.det_data, help="Observation detdata key for the timestream data"
    )

    iir_params = Unicode(
        "iir_params", help="Observation key for readout filter parameters"
    )

    wafer_key = Unicode("det_info:stream_id", help="Focalplane key for the wafer name")

    debug_root = Unicode(
        None,
        allow_none=True,
        help="Optional root filename for debug plots",
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        return

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()

        for ob in data.obs:
            if self.iir_params not in ob:
                msg = "Cannot apply readout filter. "
                msg += f"'{self.iir_params}' does not exist in '{ob.name}'"
                raise RuntimeError(msg)

            # Get the sample rate from the data.  We also have nominal sample rates
            # from the noise model and also from the focalplane.
            (rate, dt, dt_min, dt_max, dt_std) = rate_from_times(
                ob.shared[self.times].data
            )
            order = int(np.ceil(np.log(ob.n_local_samples) / np.log(2)))
            n_fft = 2 ** (order + 1)
            freq = np.fft.rfftfreq(n_fft, d=1.0 / rate)

            # Get valid local detectors
            local_dets = ob.select_local_detectors()
            local_set_dets = set(local_dets)

            # Get the rows of the focalplane table containing these dets
            det_table = ob.telescope.focalplane.detector_data
            det_to_row = {
                y: x for x, y in enumerate(det_table["name"]) if y in local_set_dets
            }
            fp_rows = np.array([det_to_row[x] for x in local_dets])

            # Get the set of all stream IDs
            all_wafers = set(det_table[self.wafer_key][fp_rows])

            # The IIR filter parameters will either be in a single,
            # top-level dictionary or they are organized per-UFM.
            if (
                "per_stream" in ob[self.iir_params]
                and ob[self.iir_params]["per_stream"]
            ):
                # We need to filter one wafer at a time.
                for wf in all_wafers:
                    wafer_dets = [
                        x
                        for x, y in zip(local_dets, det_table[self.wafer_key][fp_rows])
                        if y == wf
                    ]
                    signal = ob.detdata[self.det_data][wafer_dets, :]
                    self._filter_detectors(rate, freq, signal, ob[wf])
            else:
                # We are filtering all detectors at once
                signal = ob.detdata[self.det_data][local_dets, :]
                self._filter_detectors(rate, freq, signal, ob[self.iir_params])

    def _filter_detectors(self, rate, freq, det_array, iir_props):
        # Get the common filter kernel for all detectors
        iir_filter = filters.iir_filter(iir_params=iir_props)(freq, None)

        # Deconvolve
        convolve(
            det_array,
            rate,
            kernel_freq=freq,
            kernels=iir_filter,
            kernel_func=None,
            deconvolve=True,
            algorithm="numpy",
            debug=self.debug_root,
        )

    def _finalize(self, data, **kwargs):
        return

    def _requires(self):
        req = {
            "shared": self.times,
            "detdata": self.det_data,
        }
        return req

    def _provides(self):
        prov = {"meta": list(), "shared": list(), "detdata": list()}
        return prov

    def _accelerators(self):
        return list()



import yaml

from influxdb import InfluxDBClient

class Monitor:
    def __init__(self, host, port, database='qds', username=u'root', password=u'root', path='', ssl=False):
        """QDS Monitor, an interface to monitoring data quality in InfluxDB.

        Parameters
        ----------
        host : str
            InfluxDB host address
        port : int
            InfluxDB port number
        database : str
            InfluxDB database. Will be created if it does not exist already.
            Defaults to 'qds'.
        username : str
            Username for the InfluxDB, defaults to 'root'.
        password : str
            Password for the InfluxDB, defaults to 'root'.
        path : str
            Path of InfluxDB on the server to connect to, defaults to ''
        ssl : bool
            Use https to connect, defaults to False

        Attributes
        ----------
        client : influxdb.client.InfluxDBClient
            InfluxDB client
        queue : list
            InfluxQL line formatted entries for upload to InfluxDB. Recorded
            entries are "queued" to this list and written with Monitor.write().

        """
        self.client = Monitor._connect_to_db(host, port, database, username, password, path, ssl)
        self.queue = []

    @classmethod
    def from_configs(cls, configs):
        """Create a monitor from a configuration file
        
        Parameters
        ----------
        configs: dict or string
            configuration dictionary or string that's a file name that can be
            loaded by yaml into a configuration dictionary

        Returns
        -------
        connected Monitor Instance
        """
        if type(configs) == str:
            configs = yaml.safe_load( open(configs, "r") )    
        return cls(
            host = configs["host"],
            port = configs["port"],
            database = configs["database"],
            username = configs["username"],
            password = configs["password"],
            path = configs["path"],
            ssl = configs["ssl"],
        )

    @staticmethod
    def _connect_to_db(host, port, database, username, password, path, ssl):
        """Initailize the DB client.

        Parameters
        ----------
        host : str
            InfluxDB host address
        port : int
            InfluxDB port number
        database : str
            InfluxDB database. Will be created if it does not exist already.
        username : str
            Username for the InfluxDB.
        password : str
            Password for the InfluxDB.
        path : str
            Path of InfluxDB on the server to connect to
        ssl : bool
            Use https to connect

        Returns
        ----------
        influxdb.client.InfluxDBClient
            InfluxDB client connected to specified database

        """
        if ssl:
            verify_ssl=True
        else:
            verify_ssl=False

        client = InfluxDBClient(host=host, port=port, username=username,
                                password=password, path=path, ssl=ssl,
                                verify_ssl=verify_ssl)
        db_list = client.get_list_database()
        db_names = [x['name'] for x in db_list]
        if database not in db_names:
            print(f"{database} DB doesn't exist, creating DB")
            client.create_database(database)
        client.switch_database(database)

        return client

    def check(self, field, observation, tags, log="obs_process_log"):
        """Check if monitored measurement has been reacorded already.

        All recorded measurement fields within the Monitor are tracked in a log
        within InfluxDB. This check will search this log with a search like::

            SELECT {field} FROM "log" WHERE observation = {observation} AND
                {tag1} = '{value1}' AND {tag2} = '{value2}';

        Parameters
        ----------
        field : str
            Measurement field to check calculation for, i.e. "white_noise_level"
        observation : str
            Observation ID
        tags : dict
            Other tags to included in AND search
        log : str
            Measurement name for the log within influxdb

        Returns
        -------
        bool
            True if calculation already performed, False otherwise

        """
        query_where = f"select {field} from \"{log}\" WHERE observation = '{observation}'"

        for tag_name, tag_value in tags.items():
            and_term = f" AND {tag_name} = '{tag_value}'"
            query_where += and_term

        result = self.client.query(query_where)

        if list(result.get_points(measurement=log)):
            print(f"field {field} for observation {observation} " +
                  f"and tags {tags} already recorded in {log}")
            return True

        return False

    @staticmethod
    def _build_single_line_entry(field, value, timestamp, tags, measurement):
        """Build a single line formatted string for insertion to InfluxDB.

        Creates a string of the form:
            '{measurement},{tag}={tag_value} {field}={value} {timestamp}'

        For many tags and tag values.

        Parameters
        ----------
        field : str
            Measurement field, i.e. "white_noise_level"
        value : float or int
            Value for the field
        timestamp : float
            Timestamp for the field value (can be None, which uses time of insertion to DB)
        tags : list of dict
            List of dictionaries containing tags for the InfluxDB
        measurement : str
            InfluxDB measurement to record to

        Returns
        -------
        str
            Single InfluxDB line formatted string.

        """

        # Single value/timestamp/tags
        influxdata = f'{measurement}'

        for tag, tag_value in tags.items():
            tag_string = f',{tag}={tag_value}'
            influxdata += tag_string

        influxdata += f' {field}={value}'

        if timestamp is not None:
            time_ns = int(timestamp*1e9)
            influxdata += f' {time_ns}'

        return influxdata

    def record(self, field, values, timestamps, tags, measurement, log='obs_process_log', log_tags=None):
        """Record a monitored statistic to the InfluxDB. Values not written to
        DB until ``Monitor.write()`` is called.

        Parameters
        ----------
        field : str
            Measurement field, i.e. "white_noise_level"
        values : list or np.array
            Values for the field for each unique set of tags and timestamps
        timestamps : list or np.array
            Timestamps for the field values
        tags : list of dict
            List of dictionaries containing tags for the InfluxDB
        measurement : str
            InfluxDB measurement to record to
        log : str
            InfluxDB measurement to use for logging completed calculation
        log_tags : list of dict
            Tags to use for the log, typically you won't want to record you've
            completed a calculation for each individual detector, but maybe some higher
            level group. If this is None tags will be used.

        """
        assert len(timestamps) == len(values) == len(tags)

        # Multi values/timestamps/tags
        for (value, ts, tag_dict) in zip(values, timestamps, tags):
            data_line = Monitor._build_single_line_entry(field, value, ts, tag_dict, measurement)
            self.queue.append(data_line)

        # Log into obs_process_log measurement in InfluxDB
        if log_tags is None:
            log_tags = tags

        log_msg = Monitor._build_single_line_entry(field, 1, None, log_tags, log)
        self.queue.append(log_msg)

    def write(self):
        """Write points to InfluxDB, clearing the queue."""
        self.client.write_points(self.queue, protocol='line')
        self.queue = []



"""Functions for estimating and subtracting the temperature to polarization leakage.
"""
import numpy as np
from sotodlib import core
from sotodlib.tod_ops import filters, apodize
from sotodlib.tod_ops.fft_ops import calc_psd, calc_wn
from scipy.odr import ODR, Model, RealData
from lmfit import Model as LmfitModel


def t2p_fit(aman, T_sig_name='dsT', Q_sig_name='demodQ', U_sig_name='demodU',
            sigma_T=None, sigma_demod=None, flag_name=None, ds_factor=100):
    """

    Compute the leakage coefficients from temperature (T) to polarization (Q and U)
    individually.  Return an axismanager of coefficients with their 
    statistical uncertainties and reduced chi-squared values for the fit.

    Parameters
    ----------
    aman : AxisManager
        AxisManager object containing the TOD data.
    T_sig_name : str
        Name of the temperature signal in `aman`. Default is 'dsT'.
    Q_sig_name : str
        Name of the Q polarization signal in `aman`. Default is 'demodQ'.
    U_sig_name : str
        Name of the U polarization signal in `aman`. Default is 'demodU'.
    sigma_T : array-like
        Array of standard deviations for the signal TOD used fitting.
        If None, all are set to unity.
    sigma_demod : array-like or None
        Array of standard deviations for the Q and U TODs used as weights
        while fitting.  If None, all are set to unity.
    flag_name : str or None
        Name of the flag field in `aman` to use for masking data. If None,
        no masking is applied.
    ds_factor : float
        Factor by which to downsample the TODs prior to fitting. Default is 100. 

    Returns
    -------
    out_aman : AxisManager
        An AxisManager containing leakage coefficients, their errors, and reduced
        chi-squared statistics.
    """

    def leakage_model(params, x):
        return params[0] * x + params[1]

    if sigma_T is None:
        sigma_T = np.ones_like(aman.dets.count)

    if sigma_demod is None:
        sigma_demod = np.ones_like(aman.dets.count)

    if flag_name is None:
        mask = np.ones_like(aman[T_sig_name], dtype='bool')
    elif flag_name in aman.flags._fields.keys():
        mask = ~aman.flags[flag_name].mask()
    else:
        raise ValueError('flag_name should be in aman.flags')

    coeffsQ = np.zeros(aman.dets.count)
    errorsQ = np.zeros(aman.dets.count)
    redchi2sQ = np.zeros(aman.dets.count)
    coeffsU = np.zeros(aman.dets.count)
    errorsU = np.zeros(aman.dets.count)
    redchi2sU = np.zeros(aman.dets.count)

    for di, det in enumerate(aman.dets.vals):        
        T_ds_det = aman[T_sig_name][di][mask[di]][::ds_factor]
        Q_ds_det = aman[Q_sig_name][di][mask[di]][::ds_factor]
        U_ds_det = aman[U_sig_name][di][mask[di]][::ds_factor]

        # fitting for Q
        try:
            model = Model(leakage_model)
            data = RealData(x=T_ds_det, 
                            y=Q_ds_det, 
                            sx=np.ones_like(T_ds_det) * sigma_T[di], 
                            sy=np.ones_like(T_ds_det) * sigma_demod[di])
            odr = ODR(data, model, beta0=[np.mean(Q_ds_det), 1e-3])
            output = odr.run()
            coeffsQ[di] = output.beta[0]
            errorsQ[di] = output.sd_beta[0]
            redchi2sQ[di] = output.sum_square / (len(T_ds_det) - 2)
        except:
            coeffsQ[di] = np.nan
            errorsQ[di] = np.nan
            redchi2sQ[di] = np.nan

        # fitting for U
        try:
            model = Model(leakage_model)
            data = RealData(x=T_ds_det, 
                            y=U_ds_det, 
                            sx=np.ones_like(T_ds_det) * sigma_T[di], 
                            sy=np.ones_like(T_ds_det) * sigma_demod[di])
            odr = ODR(data, model, beta0=[np.mean(U_ds_det), 1e-3])
            output = odr.run()
            coeffsU[di] = output.beta[0]
            errorsU[di] = output.sd_beta[0]
            redchi2sU[di] = output.sum_square / (len(T_ds_det) - 2)
        except:
            coeffsU[di] = np.nan
            errorsU[di] = np.nan
            redchi2sU[di] = np.nan

    out_aman = core.AxisManager(aman.dets, aman.samps)
    out_aman.wrap('coeffsQ', coeffsQ, [(0, 'dets')])
    out_aman.wrap('errorsQ', errorsQ, [(0, 'dets')])
    out_aman.wrap('redchi2sQ', redchi2sQ, [(0, 'dets')])

    out_aman.wrap('coeffsU', coeffsU, [(0, 'dets')])
    out_aman.wrap('errorsU', errorsU, [(0, 'dets')])
    out_aman.wrap('redchi2sU', redchi2sU, [(0, 'dets')])

    return out_aman

def t2p_joint_fit(aman, T_sig_name='dsT', Q_sig_name='demodQ', U_sig_name='demodU',
                  sigma_demod=None, flag_name=None, ds_factor=100):
    """
    Compute the leakage coefficients from temperature (T) to polarization (Q and U)
    by performing a joint fit of both simultaneously. Return an axismanager of the 
    coefficients with their statistical uncertainties and reduced chi-squared values 
    for the fit.

    Parameters
    ----------
    aman : AxisManager
        AxisManager object containing the TOD data.
    T_sig_name : str
        Name of the temperature signal in `aman`. Default is 'dsT'.
    Q_sig_name : str
        Name of the Q polarization signal in `aman`. Default is 'demodQ'.
    U_sig_name : str
        Name of the U polarization signal in `aman`. Default is 'demodU'.
    sigma_demod : array-like or None
        Array of standard deviations for the Q and U TOD used as weights
        while fitting. If None, all are set to unity.
    flag_name : str or None
        Name of the flag field in `aman` to use for masking data. If None,
        no masking is applied.
    ds_factor : float
        Factor by which to downsample the TODs prior to fitting. Default is 100. 

    Returns
    -------
    out_aman : AxisManager
        An AxisManager containing leakage coefficients, their errors, and reduced
        chi-squared statistics.
    """

    def leakage_model(dT, AQ, AU, lamQ, lamU):
        return AQ + lamQ * dT + 1.j * (AU + lamU * dT)

    if sigma_demod is None:
        sigma_demod = np.ones_like(aman.dets.count)

    if flag_name is None:
        mask = np.ones_like(aman[T_sig_name], dtype='bool')
    elif flag_name in aman.flags._fields.keys():
        mask = ~aman.flags[flag_name].mask()
    else:
        raise ValueError('flag_name should be in aman.flags')

    A_Q_array = np.empty([aman.dets.count])
    A_U_array = np.empty_like(A_Q_array)
    lambda_Q_array = np.empty_like(A_Q_array)
    lambda_U_array = np.empty_like(A_Q_array)

    A_Q_error = np.empty_like(A_Q_array)
    A_U_error = np.empty_like(A_Q_array)
    lambda_Q_error = np.empty_like(A_Q_array)
    lambda_U_error = np.empty_like(A_Q_array)

    redchi2s_array = np.empty_like(A_Q_array)

    for di, det in enumerate(aman.dets.vals):
        x = aman[T_sig_name][di][mask[di]][::ds_factor]
        yQ = aman[Q_sig_name][di][mask[di]][::ds_factor]
        yU = aman[U_sig_name][di][mask[di]][::ds_factor]

        try:
            model = LmfitModel(leakage_model, independent_vars=['dT'])
            params = model.make_params(AQ=np.median(yQ), AU=np.median(yU), 
                                       lamQ=0., lamU=0.)
            result = model.fit(yQ + 1j * yU, params, dT=x,
                               weights=np.ones_like(x)/sigma_demod[di])
            A_Q_array[di] = result.params['AQ'].value
            A_U_array[di] = result.params['AU'].value
            lambda_Q_array[di] = result.params['lamQ'].value
            lambda_U_array[di] = result.params['lamU'].value

            A_Q_error[di] = result.params['AQ'].stderr
            A_U_error[di] = result.params['AU'].stderr
            lambda_Q_error[di] = result.params['lamQ'].stderr
            lambda_U_error[di] = result.params['lamU'].stderr
            redchi2s_array[di] = result.redchi
        except:
            A_Q_array[di] = np.nan
            A_U_array[di] = np.nan
            lambda_Q_array[di] = np.nan
            lambda_U_array[di] = np.nan
            
            A_Q_error[di] = np.nan
            A_U_error[di] = np.nan
            lambda_Q_array[di] = np.nan
            lambda_U_error[di] = np.nan
            redchi2s_array[di] = np.nan

    out_aman = core.AxisManager(aman.dets, aman.samps)
    out_aman.wrap('AQ', A_Q_array, [(0, 'dets')])
    out_aman.wrap('AU', A_U_array, [(0, 'dets')])
    out_aman.wrap('lamQ', lambda_Q_array, [(0, 'dets')])
    out_aman.wrap('lamU', lambda_U_array, [(0, 'dets')])
    out_aman.wrap('AQ_error', A_Q_error, [(0, 'dets')])
    out_aman.wrap('AU_error', A_U_error, [(0, 'dets')])
    out_aman.wrap('lamQ_error', lambda_Q_error, [(0, 'dets')])
    out_aman.wrap('lamU_error', lambda_U_error, [(0, 'dets')])
    out_aman.wrap('redchi2s', redchi2s_array, [(0, 'dets')])

    return out_aman

def get_t2p_coeffs(aman, T_sig_name='dsT', Q_sig_name='demodQ', U_sig_name='demodU',
                   joint_fit=True, wn_demod=None, f_lpf_cutoff=2.0, flag_name=None,
                   ds_factor=100, subtract_sig=False, merge_stats=True, 
                   t2p_stats_name='t2p_stats'):
    """
    Compute the leakage coefficients from temperature (T) to polarization (Q and U) by
    either a joint fit of both or individually. Optionally subtract this leakage. Return 
    an axismanager of the coefficients with their statistical uncertainties and reduced 
    chi-squared values for the fit.

    Parameters
    ----------
    aman : AxisManager
        AxisManager object containing the TOD data.
    T_sig_name : str
        Name of the temperature signal in `aman`. Default is 'dsT'.
    Q_sig_name : str
        Name of the Q polarization signal in `aman`. Default is 'demodQ'.
    U_sig_name : str
        Name of the U polarization signal in `aman`. Default is 'demodU'.
    joint_fit : bool
        Whether to fit Q and U leakage coefficients as parameters in a single model or
        fit independently. Default is True.
    wn_demod : float or str or None
        Precomputed white noise level for demodulated signals. If None, it will be calculated.
        If provided by a string, `aman.get(wn_demod)` is used.
    f_lpf_cutoff : float
        Cutoff frequency of low pass filter in demodulation. Used for error bar estimation by
        combination with wn_demod. Default is 2.0.
    flag_name : str
        Name of the flag field in `aman` to use for masking data. If None, no masking is applied.
    ds_factor: float or None
        Factor by which to downsample the TODs prior to fitting. If None, the low pass filter
        frequency is used to estimate the factor.
    subtract_sig : bool
        Whether to subtract the calculated leakage from the polarization signals. Default is False.
    merge_stats : bool
        Whether to merge the calculated statistics back into `aman`. Default is True.
    t2p_stats_name : str
        Name under which to wrap the output AxisManager containing statistics. Default is 't2p_stats'.

    Returns
    -------
    out_aman : AxisManager
                An AxisManager containing leakage coefficients, their errors, and reduced
                chi-squared statistics.
    """

    # get white noise level of demod for error estimation
    if isinstance(wn_demod, str):
        wn_demod = aman.get(wn_demod)
    if wn_demod is None:
        freqs, Pxx_demod, nseg = calc_psd(aman, signal=aman[Q_sig_name], merge=False,
                                          full_output=True, noverlap=0)
        wn_demod = calc_wn(aman, pxx=Pxx_demod, freqs=freqs, nseg=nseg, low_f=0.5, high_f=1.5)

    # integrate the white noise level over frequencies to get error bar of each point
    sigma_demod = wn_demod * np.sqrt(f_lpf_cutoff)
    sigma_T = sigma_demod / np.sqrt(2)

    if ds_factor is None:
        ds_factor = int(np.mean(1. / np.diff(aman.timestamps)) / (f_lpf_cutoff))

    if joint_fit:
        out_aman = t2p_joint_fit(aman, T_sig_name, Q_sig_name, U_sig_name,
                                 sigma_demod, flag_name, ds_factor)
    else:
        out_aman = t2p_fit(aman, T_sig_name, Q_sig_name, U_sig_name,
                           sigma_T, sigma_demod, flag_name, ds_factor)

    if subtract_sig:
        subtract_t2p(aman, out_aman)
    if merge_stats:
        aman.wrap(t2p_stats_name, out_aman)

    return out_aman

def subtract_t2p(aman, t2p_aman, T_signal=None):
    """
    Subtract T to P leakage.

    Parameters
    ----------
    aman : AxisManager
        The tod.
    t2p_aman : AxisManager
        Axis manager with Q and U leakage coeffients.
        If joint fitting was used in get_t2p_coeffs, Q coeffs are in 
        fields ``lamQ`` and ``AQ`` and U coeffs are in ``lamU`` and 
        ``AU``. Otherwise Q coeff is in field ``coeffsQ`` and U coeff 
        in ``coeffsU``.
    T_signal : array
        Temperature signal to scale and subtract from Q/U.
        Default is ``aman['dsT']``.
    """
    if T_signal is None:
        T_signal = aman['dsT']

    if 'AQ' in t2p_aman._assignments:
        aman.demodQ -= (T_signal * t2p_aman.lamQ[:, np.newaxis] + t2p_aman.AQ[:, np.newaxis])
        aman.demodU -= (T_signal * t2p_aman.lamU[:, np.newaxis] + t2p_aman.AU[:, np.newaxis])
    elif 'coeffsQ' in t2p_aman._assignments:
        aman.demodQ -= np.multiply(T_signal.T, t2p_aman.coeffsQ).T
        aman.demodU -= np.multiply(T_signal.T, t2p_aman.coeffsU).T
    else:
        raise ValueError('no leakage coefficients found in axis manager')



# Copyright (c) 2020 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""Demonstrate construction of some simple metadata structures.  This
includes HDF5 IO helper routines, and the ObsDb/DetDb resolution and
association system used in Context/SuperLoader.

"""

import unittest
import tempfile

from sotodlib.core import metadata
from sotodlib.io.metadata import ResultSetHdfLoader, write_dataset, _decode_array

import os
import h5py
import sqlite3

from ._helpers import mpi_multi


@unittest.skipIf(mpi_multi(), "Running with multiple MPI processes")
class MetadataTest(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.TemporaryDirectory()

    def tearDown(self):
        self.tempdir.cleanup()

    def test_000_support(self):
        """Test some numpy-HDF5 conversion support functions.

        """
        rs = metadata.ResultSet(keys=['a_string', 'a_float', 'a_bad_string', 'a_bad_float'])
        rs.rows.append(('hello', 1.2, 'yuck', 1.3))
        aru = rs.asarray(hdf_compat=True)
        self.assertTrue(aru.dtype['a_string'].char == 'S')
        # Conversion code.
        arx = _decode_array(aru, key_map={
            'a_string': 'another_string',
            'a_float': 'another_float',
            'a_bad_string': None,
            'a_bad_float': None,
            })
        self.assertEqual(list(arx.dtype.names), ['another_string', 'another_float'])
        self.assertEqual(arx['another_string'].dtype.char, 'U')

    def test_001_hdf(self):
        """Test metadata write/read to HDF5 datasets

        """
        hdf_fn = os.path.join(self.tempdir.name, '_test_000_hdf.h5')

        # The reason we're here today is that this things works but is
        # going to be removed.
        loader = ResultSetHdfLoader()
        test_obs_id = 'testobs_1234'

        # Create an hdf5 dataset which is a structured array with only the
        # 'timeconst' column, containing the single fixed value.  Since there
        # are no columns with names prefixed by 'dets:' or 'obs:', this value
        # will be broadcast to all observations and detectors that access it.
        TGOOD = 1e-3
        rs = metadata.ResultSet(keys=['timeconst'])
        rs.append({'timeconst': TGOOD})
        with h5py.File(hdf_fn, 'a') as fout:
            # Simple one...
            write_dataset(rs, fout, 'timeconst_1ms', overwrite=True)

        # Simple look-up:
        req = {'filename': hdf_fn,
               'obs:obs_id': test_obs_id,
               'dataset': 'timeconst_1ms'}
        data = loader.from_loadspec(req)
        self.assertEqual(list(data['timeconst']), [TGOOD])

    def test_010_manifest_basics(self):
        """Test that you can create a ManifestScheme and ManifestDb and add
        records but not duplicates unless you need to.

        """
        scheme = metadata.ManifestScheme() \
                         .add_range_match('obs:timestamp')
        mandb = metadata.ManifestDb(scheme=scheme)
        mandb.add_entry({'obs:timestamp': (0, 1e9)}, 'a.h5')

        # Duplicate index prevented?
        with self.assertRaises(sqlite3.IntegrityError):
            mandb.add_entry({'obs:timestamp': (0, 1e9)}, 'b.h5')

        # Replace accepted?
        mandb.add_entry({'obs:timestamp': (0, 1e9)}, 'c.h5',
                        replace=True)

        # Returns the single correct value?
        self.assertEqual(
            'c.h5', mandb.match({'obs:timestamp': 100})['filename'])

    def test_020_db_resolution(self):
        """Test metadata detdb/obsdb resolution system

        This tests one of the more complicated cases:

        - The ManifestDb includes restrictions on dets:band, so f090
          is to be loaded from one dataset and f150 is to be loaded
          from another.

        - The two datasets both provide values for f090 and f150, so
          the code has to know to ignore the ones that weren't asked
          for.

        """
        hdf_fn = os.path.join(self.tempdir.name, '_test_010_dbs.h5')
        mandb_fn = os.path.join(self.tempdir.name, '_test_010_dbs.sqlite')

        # Add two datasets to the HDF file.  They are called
        # "timeconst_early" and "timeconst_late" but there is no
        # specific time range associated with each.  Each dataset
        # contains a value for bands f090 and f150.  The "early" set
        # has TBAD for f150 and the "late" set has TBAD for f090.
        T090, T150, TBAD = 90e-3, 150e-3, 1e0
        with h5py.File(hdf_fn, 'a') as fout:
            # First test.
            for label, tau1, tau2 in [('early', T090, TBAD),
                                      ('late', TBAD, T150)]:
                rs = metadata.ResultSet(keys=['dets:band', 'timeconst'])
                rs.append({'dets:band': 'f090', 'timeconst': tau1})
                rs.append({'dets:band': 'f150', 'timeconst': tau2})
                write_dataset(rs, fout, 'timeconst_%s' % label, overwrite=True)

        # To match the early/late example we need DetDb and ObsDb.
        detdb = metadata.DetDb()
        detdb.create_table('base', ["`readout_id` str", "`band` str", "`polcode` str"])
        detdb.add_props('base', 'det1', readout_id='det1', band='f090', polcode='A')
        detdb.add_props('base', 'det2', readout_id='det2', band='f090', polcode='B')
        detdb.add_props('base', 'det3', readout_id='det3', band='f150', polcode='A')
        detdb.add_props('base', 'det4', readout_id='det4', band='f150', polcode='B')

        obsdb = metadata.ObsDb()
        t_pivot = 2000010000
        obsdb.add_obs_columns(['timestamp float'])
        obsdb.update_obs('obs_00', {'timestamp': t_pivot - 10000})
        obsdb.update_obs('obs_01', {'timestamp': t_pivot + 10000})

        # Test 1 -- ManifestDb and Stored datasets both have "band" rules.
        scheme = metadata.ManifestScheme() \
                         .add_range_match('obs:timestamp') \
                         .add_data_field('dets:band') \
                         .add_data_field('dataset')
        mandb = metadata.ManifestDb(scheme=scheme)
        for band, this_pivot in [('f090', t_pivot + 1e6),
                                 ('f150', t_pivot - 1e6)]:
            mandb.add_entry({'dataset': 'timeconst_early',
                             'dets:band': band,
                             'obs:timestamp': (0, this_pivot)},
                            filename=hdf_fn)
            mandb.add_entry({'dataset': 'timeconst_late',
                             'dets:band': band,
                             'obs:timestamp': (this_pivot, 4e9)},
                            filename=hdf_fn)
        mandb.to_file(mandb_fn)

        # The SuperLoader is where the logic lives to combine multiple
        # results and pull out the right information in the right
        # order.  It should leave us with no TBAD values.
        loader = metadata.SuperLoader(obsdb=obsdb, detdb=detdb)
        spec_list = [
            {'db': mandb_fn,
             'name': 'tau&timeconst'}
        ]
        mtod = loader.load(spec_list, {'obs:obs_id': 'obs_00'}, detdb.props())
        self.assertEqual(list(mtod['tau']), [T090, T090, T150, T150])

        # Make sure that also plays well with det specs.
        mtod = loader.load(spec_list, {'obs:obs_id': 'obs_00', 'dets:band': 'f090'},
                           detdb.props())
        self.assertEqual(list(mtod['tau']), [T090, T090])
        mtod = loader.load(spec_list, {'obs:obs_id': 'obs_00', 'dets:band': 'f150'},
                           detdb.props())
        self.assertEqual(list(mtod['tau']), [T150, T150])

        # Test 2: ManifestDb specifies polcode, which crosses with
        # dataset band.
        scheme = metadata.ManifestScheme() \
                         .add_range_match('obs:timestamp') \
                         .add_data_field('dets:polcode') \
                         .add_data_field('dataset')
        mandb = metadata.ManifestDb(scheme=scheme)
        for polcode, this_pivot in [('A', t_pivot + 1e6),
                                    ('B', t_pivot - 1e6)]:
            mandb.add_entry({'dataset': 'timeconst_early',
                             'dets:polcode': polcode,
                             'obs:timestamp': (0, this_pivot)},
                            filename=hdf_fn)
            mandb.add_entry({'dataset': 'timeconst_late',
                             'dets:polcode': polcode,
                             'obs:timestamp': (this_pivot, 4e9)},
                            filename=hdf_fn)
        mandb.to_file(mandb_fn)

        # Now we expect only f090 A and f150 B to resolve to non-bad vals.
        # Make sure you reinit the loader, to avoid cached dbs.
        loader = metadata.SuperLoader(obsdb=obsdb, detdb=detdb)
        mtod = loader.load(spec_list, {'obs:obs_id': 'obs_00'}, detdb.props())
        self.assertEqual(list(mtod['tau']), [T090, TBAD, TBAD, T150])

        # Make sure that also plays well with det specs.
        mtod = loader.load(spec_list, {'obs:obs_id': 'obs_00', 'dets:band': 'f090'},
                           detdb.props())
        self.assertEqual(list(mtod['tau']), [T090, TBAD])
        mtod = loader.load(spec_list, {'obs:obs_id': 'obs_00', 'dets:band': 'f150'},
                           detdb.props())
        self.assertEqual(list(mtod['tau']), [TBAD, T150])

    def test_030_manipulation(self):
        """Test some helpers for inspecting and updating ManifestDbs.

        """
        scheme = metadata.ManifestScheme() \
                         .add_range_match('obs:timestamp') \
                         .add_exact_match('wafer') \
                         .add_data_field('dets:band') \
                         .add_data_field('dataset')

        mandb = metadata.ManifestDb(scheme=scheme)

        mandb.add_entry({'dets:band': 'f150',
                         'wafer': 'A',
                         'obs:timestamp': (1200000000, 1300000000),
                         'dataset': 'early'}, filename='x')
        mandb.add_entry({'dets:band': 'f150',
                         'wafer': 'A',
                         'obs:timestamp': (1200000000, 1300000001),
                         'dataset': 'early'}, filename='x')
        mandb.add_entry({'dets:band': 'f220',
                         'wafer': 'A',
                         'obs:timestamp': (1300000000, 1400000001),
                         'dataset': 'early'}, filename='y')

        # Does .get_entries work?
        entries = mandb.get_entries(['obs:timestamp__lo', 'obs:timestamp__hi'])
        print(entries)

        # Does .inspect work properly?
        entries = mandb.inspect({'wafer': 'A'})
        self.assertEqual(len(entries), 3)
        self.assertIn('wafer', entries[0])

        entries = mandb.inspect({'dets:band': 'f150'})
        self.assertEqual(len(entries), 2)
        entries = mandb.inspect({'filename': 'x'})
        self.assertEqual(len(entries), 2)

        # Modify an entry
        entries[1]['wafer'] = 'B'
        entries[1]['dets:band'] = 'f090'
        entries[1].pop('filename')
        mandb.update_entry(entries[1])
        entries = mandb.inspect({'wafer': 'A'})
        self.assertEqual(len(entries), 2)

        # Delete an entry
        entries = mandb.inspect({'dets:band': 'f220'})
        mandb.remove_entry(entries[0])
        ## check file unreg'd
        c = mandb.conn.execute("select count(id) from files where name='y'")
        self.assertEqual(c.fetchall()[0][0], 0)

        # Delete another entry
        entries = mandb.inspect()
        mandb.remove_entry(entries[0])
        ## check file not unreg'd (because it's used twice)
        c = mandb.conn.execute("select count(id) from files where name='x'")
        self.assertEqual(1, c.fetchone()[0])


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2023-2023 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Simulated detector and readout effects.

This includes things fundamental to the detector and readout chain, including
timeconstant, noise, tuning yield, and readout systematics.

"""

import numpy as np
from astropy import units as u
import toast
import toast.ops

from .. import ops as so_ops
from .job import workflow_timer


def setup_simulate_detector_timeconstant(operators):
    """Add commandline args and operators for timeconstant convolution.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(
        toast.ops.TimeConstant(
            name="convolve_time_constant", deconvolve=False, enabled=False
        )
    )


@workflow_timer
def simulate_detector_timeconstant(job, otherargs, runargs, data):
    """Simulate the effects of detector timeconstants.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if otherargs.realization is not None:
        job_ops.convolve_time_constant.realization = otherargs.realization

    if job_ops.convolve_time_constant.enabled:
        job_ops.convolve_time_constant.apply(data)


def setup_simulate_detector_noise(operators):
    """Add commandline args and operators for simulating detector noise.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(toast.ops.SimNoise(name="sim_noise", enabled=False))


@workflow_timer
def simulate_detector_noise(job, otherargs, runargs, data):
    """Simulate the intrinsic detector and readout noise.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if otherargs.realization is not None:
        job_ops.sim_noise.realization = otherargs.realization

    if job_ops.sim_noise.enabled:
        job_ops.sim_noise.apply(data)


def setup_simulate_readout_effects(operators):
    """Add commandline args and operators for simulating readout effects.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(so_ops.SimReadout(name="sim_readout", enabled=False))


@workflow_timer
def simulate_readout_effects(job, otherargs, runargs, data):
    """Simulate various readout effects.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.sim_readout.enabled:
        job_ops.sim_readout.apply(data)


def setup_simulate_detector_yield(operators):
    """Add commandline args and operators for simulating detector yield.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(toast.ops.YieldCut(name="yield_cut", enabled=False))


@workflow_timer
def simulate_detector_yield(job, otherargs, runargs, data):
    """Simulate detector yield.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if job_ops.yield_cut.enabled:
        job_ops.yield_cut.apply(data)


def setup_simulate_mumux_crosstalk(operators):
    """Add commandline args and operators for simulating nonlinear muMUX crosstalk.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(so_ops.SimMuMUXCrosstalk(name="sim_mumux_crosstalk", enabled=False))


@workflow_timer
def simulate_mumux_crosstalk(job, otherargs, runargs, data):
    """Simulate nonlinear muMUX crosstalk.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    # Configured operators for this job
    job_ops = job.operators

    if otherargs.realization is not None:
        job_ops.sim_mumux_crosstalk.realization = otherargs.realization

    if job_ops.sim_mumux_crosstalk.enabled:
        job_ops.sim_mumux_crosstalk.apply(data)



# Copyright (c) 2022 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

"""Test Context class, including metadata and TOD loading.

"""

import unittest
import tempfile

from sotodlib import core
from sotodlib.core import metadata, Context, OBSLOADER_REGISTRY
from sotodlib.io.metadata import ResultSetHdfLoader, write_dataset, _decode_array
import so3g

import numpy as np
import os
import h5py
import sqlite3
import yaml

from ._helpers import mpi_multi

MINIMAL_CONTEXT = {
    'tags': {},
    'imports': [],
    'metadata': [],
    }


@unittest.skipIf(mpi_multi(), "Running with multiple MPI processes")
class ContextTest(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.TemporaryDirectory()

    def tearDown(self):
        self.tempdir.cleanup()

    def _open_new(self, name, mode='w'):
        full = os.path.join(self.tempdir.name, name)
        if mode == 'create':
            open(full, 'w').close()
            fout = None
        else:
            fout = open(full, mode)
        return full, fout

    def _write_context(self, d):
        name, fout = self._open_new('context.yaml')
        fout.write(yaml.dump(d))
        fout.close()
        return name

    def test_000_smoke(self):
        """Test basic instantiation of Context and basic DBs from files."""
        ctx_file = self._write_context(MINIMAL_CONTEXT)
        ctx = Context(ctx_file)
        metadata.cli.main(args=['context', ctx_file])
        del ctx, ctx_file

        # Does context populate obsdb, detdb, obsfiledb from file?
        for key, cls in [
                ('obsdb', metadata.ObsDb),
                ('detdb', metadata.DetDb),
                ('obsfiledb', metadata.ObsFileDb),
        ]:
            cd = MINIMAL_CONTEXT.copy()
            db = cls()
            name, _ = self._open_new('db.sqlite', 'create')
            db.to_file(name)
            cd[key] = name
            ctx_file = self._write_context(cd)
            ctx = Context(ctx_file)
            self.assertIsInstance(getattr(ctx, key), cls,
                                  msg=f"Instantiating '{key}'")
            metadata.cli.main(args=['context', ctx_file])

    def test_010_cli(self):
        def save_db(src, k, suffix=''):
            if src.get(k) is None:
                return
            if isinstance(src[k], str):
                return
            db_file, _ = self._open_new(k + suffix, 'create')
            src[k].to_file(db_file)
            src[k] = db_file

        dataset_sim = DatasetSim()
        ctx = dataset_sim.get_context()
        for _db in ['obsfiledb', 'obsdb', 'detdb']:
            print(ctx)
            save_db(ctx, _db)
        for i, entry in enumerate(ctx.get('metadata', [])):
            save_db(entry, 'db', str(i))

        ctx_file = self._write_context(dict(ctx))
        metadata.cli.main(args=['context', ctx_file])

    def test_100_loads(self):
        dataset_sim = DatasetSim()

        # Test detector resolutions ...
        ctx = dataset_sim.get_context()
        n = len(dataset_sim.dets)
        obs_id = dataset_sim.obss['obs_id'][1]
        for selection, count in [
                (['read05'], 1),
                (np.array(['read05']), 1),
                (metadata.ResultSet(['readout_id'], [['read05']]), 1),
                ({'dets:readout_id': ['read05']}, 1),
                ({'dets:readout_id': np.array(['read05'])}, 1),
                ({'dets:detset': 'neard'}, 4),
                ({'dets:detset': ['neard']}, 4),
                ({'dets:detset': ['neard', 'fard']}, 8),
                ({'dets:detset': ['neard'], 'dets:readout_id': ['read00', 'read05']}, 1),
                ({'dets:band': ['f090']}, 4),
                ({'dets:band': ['f090'], 'dets:detset': ['neard']}, 2),
                ({'dets:det_id': ['NO_MATCH']}, 3),
        ]:
            meta = ctx.get_meta(obs_id, dets=selection)
            self.assertEqual(meta.dets.count, count, msg=f"{selection}")
            self.assertTrue('cal' in meta)
            self.assertTrue('flags' in meta)
            self.assertTrue('freeform' in meta)
            self.assertTrue('samps_only' in meta)

        # And tolerance of the detsets argument ...
        for selection, count in [
                ('neard', 4),
                (['neard'], 4),
                (np.array(['neard', 'fard']), 8),
        ]:
            meta = ctx.get_meta(obs_id, detsets=selection)
            self.assertEqual(meta.dets.count, count, msg=f"{selection}")
            self.assertTrue('cal' in meta)
            self.assertTrue('flags' in meta)
            self.assertTrue('freeform' in meta)
            self.assertTrue('samps_only' in meta)

        # And without detdb nor metadata
        ctx = dataset_sim.get_context(with_detdb=False)
        for selection, count in [
                ({'dets:detset': ['neard']}, 4),
        ]:
            meta = ctx.get_meta(obs_id, dets=selection)
            self.assertEqual(meta.dets.count, count, msg=f"{selection}")

        # Using free tags
        ctx = dataset_sim.get_context()
        for _id, _tags, count in [
                (obs_id, ['f090'], 4),
                (obs_id + ':f090', [], 4),
        ]:
            meta = ctx.get_meta(_id, free_tags=_tags)
            self.assertEqual(meta.dets.count, count, msg=f"{selection}")

        # And with obs_id as a dict.
        meta = ctx.get_meta({'obs_id': obs_id})
        self.assertEqual(meta.dets.count, 8)

        # You know, the kind of dict you get from obsdb.
        meta = ctx.get_meta(ctx.obsdb.get()[0])
        self.assertEqual(meta.dets.count, 8)

        # Check check mode
        checks = ctx.get_meta(obs_id, check=True)
        self.assertIsInstance(checks, list)

        # Check det_info mode
        det_info = ctx.get_det_info(obs_id)
        self.assertIsInstance(det_info, metadata.ResultSet)
        self.assertEqual(len(det_info), n)

        # Check tolerant mode
        ctx = dataset_sim.get_context(with_bad_metadata=True)
        with self.assertRaises(Exception):
            ctx.get_meta(obs_id)
        ctx.get_meta(obs_id, ignore_missing=True)

        # Check det_info with acceptable partial coverage.
        ctx = dataset_sim.get_context(with_incomplete_det_info='trim')
        meta = ctx.get_meta(obs_id)
        self.assertTrue('newcal' in meta.det_info)
        self.assertLess(meta.dets.count, 8)

        ctx = dataset_sim.get_context(with_incomplete_det_info='skip')
        meta = ctx.get_meta(obs_id)
        self.assertFalse('newcal' in meta.det_info)
        self.assertEqual(meta.dets.count, 8)

        ctx = dataset_sim.get_context(with_incomplete_det_info='trim',
                                      with_dependent_metadata='trim')
        meta = ctx.get_meta(obs_id)
        self.assertTrue('depends' in meta)

        # ... but skip should fail if dependent metadata requires what
        # it is providing.
        for dep in ['fail', 'skip', 'trim']:
            # arguably this could be asked to not raise, for dep =
            # skip or trim.
            ctx = dataset_sim.get_context(with_incomplete_det_info='skip',
                                          with_dependent_metadata=dep)
            with self.assertRaises(metadata.loader.IncompleteDetInfoError):
                ctx.get_meta(obs_id)

        # Check det_info with unacceptable partial coverage.
        ctx = dataset_sim.get_context(with_incomplete_det_info='fail')
        with self.assertRaises(metadata.loader.IncompleteMetadataError):
            ctx.get_meta(obs_id)

        # Manual overrides.
        meta = ctx.get_meta(obs_id, on_missing={'newcal': 'trim'})
        self.assertTrue('newcal' in meta.det_info)
        self.assertLess(meta.dets.count, 8)

        meta = ctx.get_meta(obs_id, on_missing={'newcal': 'skip'})
        self.assertFalse('newcal' in meta.det_info)
        self.assertEqual(meta.dets.count, 8)

        # Check metadata with acceptable partial coverage.
        ctx = dataset_sim.get_context(with_incomplete_metadata='trim')
        meta = ctx.get_meta(obs_id)
        self.assertTrue('othercal' in meta)
        self.assertLess(meta.dets.count, 8)

        dataset_sim = DatasetSim()
        obs_id = dataset_sim.obss['obs_id'][1]

        ctx = dataset_sim.get_context(with_incomplete_metadata='skip')
        meta = ctx.get_meta(obs_id)
        self.assertEqual(meta.dets.count, 8)
        self.assertFalse('othercal' in meta)

        # Check metadata with unacceptable partial coverage.
        ctx = dataset_sim.get_context(with_incomplete_metadata='fail')
        with self.assertRaises(metadata.loader.IncompleteMetadataError):
            ctx.get_meta(obs_id)

        # Manual overrides.
        meta = ctx.get_meta(obs_id, on_missing={'othercal': 'trim'})
        self.assertTrue('othercal' in meta)
        self.assertLess(meta.dets.count, 8)

        meta = ctx.get_meta(obs_id, on_missing={'othercal': 'skip'})
        self.assertEqual(meta.dets.count, 8)
        self.assertFalse('othercal' in meta)

        # Nothing wrong with good old obs 13 though
        ctx.get_meta('obs_number_13')

    def test_110_more_loads(self):
        dataset_sim = DatasetSim()
        n_det, n_samp = dataset_sim.det_count, dataset_sim.sample_count
        obs_id = dataset_sim.obss['obs_id'][1]

        # Test some TOD loading.
        ctx = dataset_sim.get_context()
        tod = ctx.get_obs(obs_id)
        self.assertEqual(tod.signal.shape, (n_det, n_samp))

        # ... metadata reconciliation check
        self.assertTrue(np.all(tod.cal > 0))
        for band, f in zip(tod.det_info['band'], tod.flags):
            # The f090 dets should have 0 flag intervals; f150 have 1
            self.assertEqual(len(f.ranges()), int(band == 'f150'))
        # Check if NO_MATCH det_id seemed to broadcast propertly ...
        self.assertEqual(list(tod.det_info['det_param'] == -1),
                         list(dataset_sim.dets['det_id'] == 'NO_MATCH'))
        for di, ds, AB1, AB2 in zip(tod.det_info.det_id, tod.det_info.detset,
                                    tod.XY, tod.focal_plane2.AB):
            self.assertEqual(AB1, len(ds) * (di == 'NO_MATCH'))
            self.assertEqual(AB2, len(ds) * (di == 'NO_MATCH'))

        tod = ctx.get_obs(obs_id + ':f090')
        self.assertEqual(tod.signal.shape, (n_det // 2, n_samp))

        tod = ctx.get_obs(obs_id, free_tags=['f150'])
        self.assertEqual(tod.signal.shape, (n_det // 2, n_samp))

        tod = ctx.get_obs(obs_id, samples=(10, n_samp // 2))
        self.assertEqual(tod.signal.shape, (n_det, n_samp // 2 - 10))

        # Loading via filename
        tod = ctx.get_obs(filename='obs_number_11_neard.txt')
        self.assertEqual(tod.signal.shape, (n_det // 2, n_samp))

        # Loading via prepopulated & modified meta
        meta = ctx.get_meta(obs_id)
        meta.restrict('dets', meta.dets.vals[:5])
        tod = ctx.get_obs(obs_id=meta)
        self.assertEqual(tod.signal.shape, (5, n_samp))
        tod = ctx.get_obs(meta=meta)
        self.assertEqual(tod.signal.shape, (5, n_samp))

        meta = ctx.get_meta(obs_id, samples=(10,90))
        meta.restrict('dets', meta.dets.vals[:5])
        tod = ctx.get_obs(meta)
        self.assertEqual(tod.signal.shape, (5, 80))

        det_info = ctx.get_det_info(obs_id)
        det_info = det_info.subset(rows=det_info['band'] == 'f090')
        tod = ctx.get_obs(obs_id, dets=det_info)
        self.assertEqual(tod.signal.shape, (n_det // 2, n_samp))

        context_with_metadata = {
            'tags': {},
            'imports': [],
            'metadata': [
                {'db': 'test0.sqlite', 'label': 'test0'},
                {'db': 'test1.sqlite', 'label': 'test1'},
                {'db': 'test2.sqlite', 'label': 'test2'},
            ],
        }
        ctx_file = self._write_context(context_with_metadata)
        ctx = Context(ctx_file)
        self.assertEqual(len(ctx['metadata']), 3)
        ctx = Context(ctx_file, metadata_list=['test0'])
        self.assertEqual(len(ctx['metadata']), 1)

    def test_120_load_fields(self):
        dataset_sim = DatasetSim()
        obs_id = dataset_sim.obss['obs_id'][1]
        ctx = dataset_sim.get_context(with_axisman_ondisk=True)
        tod = ctx.get_obs(obs_id)
        self.assertCountEqual(tod.ondisk._fields.keys(), ['disk1', 'subaman'])
        self.assertCountEqual(tod.ondisk.subaman._fields.keys(), ['disk2'])

    def test_load_fields_resultset(self):
        dataset_sim = DatasetSim()
        obs_id = dataset_sim.obss['obs_id'][1]
        ctx = dataset_sim.get_context(with_resultset_ondisk=True)
        tod = ctx.get_obs(obs_id)
        # Make sure that 'band1' was loaded  into det_info
        self.assertTrue('band1' in tod.det_info._fields)
        # Make sure that 'band1' laods the same informaton as 'band'
        self.assertTrue((tod.det_info.band1 == tod.det_info.band).all())
        # Make sure that only 'pol_code' was loaded into ondisc_resultset
        self.assertCountEqual(tod.ondisk_resultset._fields.keys(), ['pol_code'])
        self.assertTrue((tod.det_info.pol_code == tod.ondisk_resultset.pol_code).all())

    def test_200_load_metadata(self):
        """Test the simple metadata load wrapper."""
        dataset_sim = DatasetSim()
        obs_id = dataset_sim.obss['obs_id'][1]

        ctx = dataset_sim.get_context()
        tod = ctx.get_meta(obs_id)
        for spec in ctx['metadata']:
            item = metadata.loader.load_metadata(tod, spec)
            assert(item is not None)
            item = metadata.loader.load_metadata(tod, spec, unpack=True)
            assert(item is not None)

    def test_300_clean_concat(self):
        """Test we can dodge concat failures."""
        dataset_sim = DatasetSim()
        obs_id = dataset_sim.obss['obs_id'][1]

        ctx = dataset_sim.get_context(with_inconcatable=True)
        m_entry = ctx['metadata'][0]
        orig_unpack = m_entry['unpack']

        # The discrepant fields should cause this to fail.
        with self.assertRaises(ValueError):
            tod = ctx.get_meta(obs_id)

        m_entry['drop_fields'] = ['discrepant_s', 'discrepant_v']
        tod = ctx.get_meta(obs_id)

        m_entry['drop_fields'] = ['discrepant_*']
        tod = ctx.get_meta(obs_id)

        m_entry['drop_fields'] = 'discrepant_*'
        tod = ctx.get_meta(obs_id)


class DatasetSim:
    """Provide in-RAM Context objects and tod/metadata loader functions
    for Context behavior testing.

    """
    def __init__(self):
        self.dets = metadata.ResultSet(
            ['readout_id', 'band', 'pol_code', 'x', 'y', 'detset', 'det_id', 'det_param'],
            [('read00', 'f090', 'A', 0.0, 0.0, 'neard', 'det00', 120.),
             ('read01', 'f090', 'B', 0.0, 0.0, 'neard', 'det01', 121.),
             ('read02', 'f150', 'A', 0.0, 0.0, 'neard', 'NO_MATCH', -1.),
             ('read03', 'f150', 'B', 0.0, 0.0, 'neard', 'NO_MATCH', -1.),
             ('read04', 'f090', 'A', 1.0, 0.0, 'fard',  'det04', 124.), 
             ('read05', 'f090', 'B', 1.0, 0.0, 'fard',  'det05', 125.),
             ('read06', 'f150', 'A', 1.0, 0.0, 'fard',  'NO_MATCH', -1.),
             ('read07', 'f150', 'B', 1.0, 0.0, 'fard',  'det07', 127.),
            ])

        self.obss = metadata.ResultSet(
            ['obs_id', 'timestamp', 'type', 'target'],
            [('obs_number_11', 1600010000., 'planet', 'uranus'),
             ('obs_number_12', 1600020000., 'survey', 'the_cmb'),
             ('obs_number_13', 1600030000., 'survey', 'the_cmb'),
            ])

        self.det_count = len(self.dets)
        self.sample_count = 100

        class _TestML(metadata.LoaderInterface):
            def from_loadspec(_self, load_params, **load_kwargs):
                return self.metadata_loader(load_params, **load_kwargs)

        OBSLOADER_REGISTRY['unittest_loader'] = self.tod_loader
        metadata.SuperLoader.register_metadata('unittest_loader', _TestML)

    def get_context(self, with_detdb=True, with_metadata=True,
                    with_bad_metadata=False, with_incomplete_det_info=False,
                    with_dependent_metadata=False,
                    with_incomplete_metadata=False,
                    with_inconcatable=False,
                    with_axisman_ondisk=False,
                    with_resultset_ondisk=False,
                    on_missing='trim'):
        """Args:
          with_detdb: if False, no detdb is included.
          with_metadata: if False, no metadata are included.
          with_bad_metadata: if True, an entry that refers to a
            non-existant sqlite database is included.
          with_incomplete_det_info: if True, include det_info entries
            that are missing some dets, or a complete detset.
          with_incomplete_metadata: if True, include entries that are
            missing some dets, or a complete detset.

        """
        detdb = metadata.DetDb()
        detdb.create_table('base', ['readout_id string',
                                    'pol_code string',
                                    'x float',
                                    'y float'])
        save_for_later = ['band', 'detset', 'det_id', 'det_param']
        for row in self.dets.subset(keys=[k for k in self.dets.keys
                                          if k not in save_for_later]):
            detdb.add_props('base', row['readout_id'], **row)

        obsdb = metadata.ObsDb()
        obsdb.add_obs_columns([f'{k} string' for k in self.obss.keys
                               if k not in ['obs_id', 'timestamp']])
        for row in self.obss:
            obsdb.update_obs(row['obs_id'], data=row)

        obsfiledb = metadata.ObsFileDb()
        for ds in ['neard', 'fard']:
            obsfiledb.add_detset(ds, self.dets['readout_id'][self.dets['detset'] == ds])
            for obs_id in self.obss['obs_id']:
                obsfiledb.add_obsfile(f'{obs_id}_{ds}.txt', obs_id, ds,
                                      0, self.sample_count)

        ctx = Context(data=MINIMAL_CONTEXT)
        ctx.obsdb = obsdb
        ctx.obsfiledb = obsfiledb
        if with_detdb:
            ctx.detdb = detdb
        ctx.reload(['loader'])
        ctx.update({
            'obs_loader_type': 'unittest_loader',
            'obs_colon_tags': ['band'],
        })

        if not with_metadata:
            return ctx

        class _ManifestDb(metadata.ManifestDb):
            def __init__(self, filename='unnamed', **kwargs):
                super().__init__(**kwargs)
                self._filename = filename
            def __repr__(self):
                return f'ManifestDb[{self._filename}]'

        def _db_single_dataset(filename):
            # Many dbs map all obs to a single file, based on
            # timestamp I guess.
            scheme = metadata.ManifestScheme() \
                             .add_data_field('loader') \
                             .add_range_match('obs:timestamp')
            db = _ManifestDb(scheme=scheme, filename=filename)
            db.add_entry(
                {'obs:timestamp': [0, 2e9], 'loader': 'unittest_loader'},
                filename)
            return db

        def _db_multi_dataset(filename, detsets=['neard', 'fard']):
            # ... while others are to a single file but by detset, using dataset=detset data arg.
            scheme = metadata.ManifestScheme() \
                      .add_data_field('dataset') \
                      .add_exact_match('dets:detset') \
                      .add_data_field('loader')
            db = _ManifestDb(scheme=scheme, filename=filename)
            for detset in detsets:
                db.add_entry(
                    {'loader': 'unittest_loader',
                     'dataset': detset,
                     'dets:detset': detset,
                     }, filename)
            return db

        # metadata: abscal.h5
        _scheme = metadata.ManifestScheme() \
                  .add_range_match('obs:timestamp') \
                  .add_data_field('loader') \
                  .add_data_field('dataset') \
                  .add_data_field('dets:band')
        abscal_db = _ManifestDb(scheme=_scheme, filename='abscal.h5')
        for band in ['f090', 'f150']:
            abscal_db.add_entry(
                {'obs:timestamp': [0, 2e9],
                 'loader': 'unittest_loader',
                 'dataset': f'mostly_{band}',
                 'dets:band': band
                },
                'abscal.h5')

        # metadata: some_flags.g3
        _scheme = metadata.ManifestScheme() \
                  .add_range_match('obs:timestamp') \
                  .add_data_field('loader') \
                  .add_data_field('frame_index', 'int') \
                  .add_data_field('dets:band')
        flags_db = _ManifestDb(scheme=_scheme, filename='some_flags.g3')
        for frame, band in enumerate(['f090', 'f150', 'f220']):
            flags_db.add_entry(
                {'obs:timestamp': [0, 2e9],
                 'loader': 'unittest_loader',
                 'frame_index': frame,
                 'dets:band': band
                 }, 'some_flags.g3')

        # metadata into context.
        ctx['metadata'] = [
            {'db': _db_single_dataset('bands.h5'),
             'det_info': True},
            {'db': _db_single_dataset('det_id.h5'),
             'det_info': True},
            {'db': abscal_db,
             'name': 'cal&abscal',
             'on_missing': on_missing},
            {'db': flags_db,
             'name': 'flags&',
             'on_missing': on_missing},
            {'db': _db_multi_dataset('some_detset_info.h5'),
             'name': 'focal_plane',
             'on_missing': on_missing},
            {'db': _db_multi_dataset('more_detset_info.h5'),
             'label': 'focal_plane2',
             'unpack': [
                 'focal_plane2',
                 'XY&AB',
                 ],
             'on_missing': on_missing},
            {'db': _db_single_dataset('freeform_info.h5'),
             'label': 'freeform',
             'unpack': 'freeform'},
            {'db': _db_single_dataset('samps_only.h5'),
             'label': 'samps_only',
             'unpack': 'samps_only'},
            {'db': _db_single_dataset('det_param.h5'),
             'det_info': True},
            {'db': _db_multi_dataset('detinfo_multimatch.h5'),
             'det_info': True},
            {'db': _db_multi_dataset('detinfo_nomatch.h5'),
             'det_info': True},
        ]

        if with_bad_metadata:
            # This entry is intended to cause a lookup failure.
            ctx['metadata'].insert(0, {
                'db': 'not-a-file.sqlite',
                'name': 'important_info&',
            })

        if with_incomplete_det_info:
            # det_info -- incomplete dataset; for key obs only a few
            ## dets will have field defined.
            ctx['metadata'].insert(0, {
                'db': _db_single_dataset('incomplete_det_info.h5'),
                'det_info': True,
                'on_missing': with_incomplete_det_info,
                'label': 'newcal',
            })

            if with_dependent_metadata:
                # metadata -- loads in association with det_info
                # fields defined through incomplete_det_info.
                ctx['metadata'].insert(1, {
                    'db': _db_single_dataset('dependent_metadata.h5'),
                    'on_missing': with_dependent_metadata,
                    'name': 'depends',
                })

        if with_incomplete_metadata:
            # metadata: incomplete_metadata.h5
            ## This is an incomplete dataset.
            _scheme = metadata.ManifestScheme() \
                      .add_exact_match('obs:obs_id') \
                      .add_data_field('loader')
            bad_meta_db = metadata.ManifestDb(scheme=_scheme)
            bad_meta_db.add_entry(
                {'obs:obs_id': 'obs_number_12', 'loader': 'unittest_loader'},
                 'incomplete_metadata.h5'
            )
            bad_meta_db.add_entry(
                {'obs:obs_id': 'obs_number_13', 'loader': 'unittest_loader'},
                 'incomplete_metadata.h5'
            )
            # This entry has incomplete metadata, which will cause a
            # failure, a trim, or omission of the product depending on
            # the value of with_incomplete_metadata.
            ctx['metadata'].insert(0, {
                'db': bad_meta_db,
                'name': 'othercal',
                'on_missing': with_incomplete_metadata,
                'label': 'othercal',
            })

        if with_inconcatable:
            _scheme = metadata.ManifestScheme() \
                      .add_exact_match('obs:obs_id') \
                      .add_exact_match('dets:detset') \
                      .add_data_field('loader')
            inconcat_db = metadata.ManifestDb(scheme=_scheme)
            for detset in ['neard', 'fard']:
                inconcat_db.add_entry(
                    {'obs:obs_id': 'obs_number_12',
                     'dets:detset': detset,
                     'loader': 'unittest_loader'},
                     'inconcat_metadata.h5'
                )
            ctx['metadata'].insert(0, {
                'db': inconcat_db,
                'unpack': ['inconcat&per_det'],
                'label': 'inconcat',
            })

        if with_axisman_ondisk:
            # Note this uses standard HDF5 loader, not our in-RAM loader.
            _scheme = metadata.ManifestScheme() \
                .add_exact_match('obs:obs_id') \
                .add_data_field('dataset')
            ondisk_db = metadata.ManifestDb(scheme=_scheme)
            ondisk_db._tempdir = tempfile.TemporaryDirectory()
            filename = os.path.join(ondisk_db._tempdir.name,
                                    'ondisk_metadata.h5')
            ondisk_db.add_entry(
                {'obs:obs_id': 'obs_number_12',
                 'dataset': 'xyz'},
                filename)
            ctx['metadata'].insert(0, {
                'db': ondisk_db,
                'unpack': 'ondisk',
                'load_fields': ['disk1', 'subaman.disk2', 'subaman.disk3'],
            })
            # Write the result.
            output = core.AxisManager(
                core.LabelAxis('dets', self.dets['readout_id']),
                core.OffsetAxis('samps', self.sample_count, 0))
            output.wrap('subaman', core.AxisManager(output.dets))
            for i in [1, 2]:
                output.wrap_new(f'disk{i}', shape=('samps',))
                output['subaman'].wrap_new(f'disk{i}', shape=('dets',))
            output.save(filename, 'xyz')

        if with_resultset_ondisk:
            _scheme = metadata.ManifestScheme() \
                .add_exact_match('obs:obs_id') \
                .add_data_field('dataset')
            ondisk_db = metadata.ManifestDb(scheme=_scheme)
            ondisk_db._tempdir = tempfile.TemporaryDirectory()
            filename = os.path.join(ondisk_db._tempdir.name,
                                    'ondisk_resultset_metadata.h5')
            write_dataset(self.dets, filename, 'obs_number_12')
            ondisk_db.add_entry(
                {'obs:obs_id': 'obs_number_12',
                 'dataset': 'obs_number_12'},
                filename)
            ctx['metadata'].insert(0, {
                'db': ondisk_db,
                'det_info': True,
                'load_fields': [{'band': 'dets:band1'}, {'readout_id': 'dets:readout_id'}],
            })

            ctx['metadata'].insert(0, {
                'db': ondisk_db,
                'unpack': 'ondisk_resultset',
                'load_fields': [{'readout_id': 'dets:readout_id'}, 'pol_code'],
            })

        return ctx


    def metadata_loader(self, kw, **load_kwargs):
        # For Superloader.
        filename = os.path.split(kw['filename'])[1]
        if filename == 'bands.h5':
            rs = self.dets.subset(keys=['readout_id', 'band'])
            rs.keys = ['dets:' + k for k in rs.keys]
            return rs
        elif filename == 'det_id.h5':
            rs = self.dets.subset(keys=['readout_id', 'det_id'])
            rs.keys = ['dets:' + k for k in rs.keys]
            return rs
        elif filename == 'det_param.h5':
            rs = self.dets.subset(keys=['det_id', 'det_param'])
            # Keep only 1 row with 'NO_MATCH'
            while sum(rs['det_id'] == 'NO_MATCH') > 1:
                rs.rows.pop(list(rs['det_id']).index('NO_MATCH'))
            rs.keys = ['dets:' + k for k in rs.keys]
            return rs
        elif filename == 'abscal.h5':
            rs = metadata.ResultSet(['dets:band', 'abscal'])
            if kw['dataset'] == 'mostly_f090':
                rs.append({'dets:band': 'f090',
                           'abscal': 90.})
                rs.append({'dets:band': 'f150',
                           'abscal': -1.})
            else:
                rs.append({'dets:band': 'f090',
                           'abscal': -1.})
                rs.append({'dets:band': 'f150',
                           'abscal': 150.})
            return rs
        elif filename == 'some_flags.g3':
            output = core.AxisManager(
                core.LabelAxis('dets', self.dets['readout_id']),
                core.OffsetAxis('samps', self.sample_count, 0))
            if kw['frame_index'] == 0:
                # Frame 0 declares zeros for all dets, but is
                # restricted by the ManifestDb to only apply to f090.
                output.wrap('flags', so3g.proj.RangesMatrix.zeros(output.shape),
                            [(0, 'dets'), (1, 'samps')])
            elif kw['frame_index'] == 1:
                # Frame 1 declares ones for only the f150 dets (and is
                # marked to apply to f150 in the ManifestDb).
                output.wrap('flags', so3g.proj.RangesMatrix.ones(output.shape),
                            [(0, 'dets'), (1, 'samps')])
                output.restrict('dets',
                                self.dets['readout_id'][self.dets['band'] == 'f150'])
            else:
                # Frame 2 is marked in the ManifestDb as being for
                # f220; the det_info preprocessing should prevent this
                # from ever getting requested.
                raise RuntimeError('metadata system asked for f220 data')
            return output
        elif filename == 'freeform_info.h5':
            output = core.AxisManager()
            output.wrap('number1', 1.)
            output.wrap('numbers', np.array([1,2,3]))
            return output
        elif filename == 'samps_only.h5':
            output = core.AxisManager(
                core.OffsetAxis('samps', self.sample_count, 0))
            output.wrap_new('encoder_something', shape=('samps', ))[:] = \
                np.arange(self.sample_count)
            return output
        elif filename == 'some_detset_info.h5':
            rs = metadata.ResultSet(['dets:readout_id', 'x', 'y'])
            for row in self.dets.subset(rows=self.dets['detset'] == kw['dets:detset']):
                rs.append({'dets:readout_id': row['readout_id'],
                           'x': 100., 'y': 102.})
            return rs
        elif filename == 'more_detset_info.h5':
            # Here, match against det_id and makes sure there's only
            # one NO_MATCH entry _per detset_.  AB will be 0 unless
            # NO_MATCH, in which case it's the len of the detset name.
            rs = metadata.ResultSet(['dets:det_id', 'AB'])
            for row in self.dets.subset(rows=self.dets['detset'] == kw['dets:detset']):
                rs.append({'dets:det_id': row['det_id'],
                           'AB': len(kw['dets:detset']) * (row['det_id'] == 'NO_MATCH')})
            while sum(rs['dets:det_id'] == 'NO_MATCH') > 1:
                rs.rows.pop(list(rs['dets:det_id']).index('NO_MATCH'))
            return rs
        elif filename == 'detinfo_multimatch.h5':
            # This is to test whether det_info fields (bp_code) can be
            # populated based on matching against multiple other
            # det_info fields (band, pol_code).
            rs = metadata.ResultSet(
                ['dets:band', 'dets:pol_code', 'dets:bp_code'], [
                    ['f090', 'A', 'f090-A'],
                    ['f090', 'B', 'f090-B'],
                    ['f150', 'A', 'f150-A'],
                    ['f150', 'B', 'f150-B'],
                ])
            return rs
        elif filename == 'detinfo_nomatch.h5':
            # This is to test that det_info can reconcile against
            # multiple det_id=NO_MATCH entries.  So at least one of
            # the detsets should have multiple det_id=NO_MATCH
            # entries.
            farness = 100. if kw['dataset'] == 'fard' else 0.001
            rs = metadata.ResultSet(['dets:det_id', 'dets:farness'])
            for row in self.dets.subset(rows=self.dets['detset'] == kw['dataset']):
                rs.append({'dets:det_id': row['det_id'],
                           'dets:farness': farness})
            return rs
        elif filename == 'incomplete_det_info.h5':
            rs = metadata.ResultSet(['dets:readout_id', 'dets:newcal'])
            for row in self.dets.subset(rows=[0,1,2]):
                rs.append({'dets:readout_id': row['readout_id'],
                           'dets:newcal':20})
            return rs
        elif filename == 'dependent_metadata.h5':
            rs = metadata.ResultSet(['dets:newcal', 'newcal_number'])
            rs.append({'dets:newcal': 20,
                       'newcal_number': 'twenty'})
            return rs
        elif filename == 'incomplete_metadata.h5':
            rs = metadata.ResultSet([
                'obs:obs_id',
                'dets:readout_id',
                'othercal'
            ])
            if kw['obs:obs_id']=='obs_number_12':
                for row in self.dets.subset(rows=[0,1,2]):
                    rs.append({'obs:obs_id': 'obs_number_12',
                               'dets:readout_id': row['readout_id'],
                               'othercal':40})
            elif kw['obs:obs_id'] =='obs_number_13':
                for row in self.dets:
                    rs.append({'obs:obs_id': 'obs_number_13',
                               'dets:readout_id': row['readout_id'],
                               'othercal':40})
            return rs

        elif filename == 'inconcat_metadata.h5':
            ds = self.dets.subset(rows=self.dets['detset'] == kw['dets:detset'])
            output = core.AxisManager(
                core.LabelAxis('dets', ds['readout_id']),
                core.OffsetAxis('samps', self.sample_count))
            # These are ok ...
            output.wrap_new('per_det', ('dets',), dtype=int)[:] = \
                np.arange(len(ds))
            output.wrap_new('duplicated', (100,))[:] = 100.
            # But put a nan in there ...
            output['duplicated'][10] = np.nan
            output.wrap('same_nan', np.nan)
            # And these are not concatenable ...
            discrepancy = int(kw['dets:detset'] == 'neard')
            output.wrap('discrepant_s', discrepancy)
            output.wrap_new('discrepant_v', (100,))[:] = discrepancy
            return output

        raise ValueError(f'metadata request for "{filename}"')

    def tod_loader(self, obsfiledb, obs_id, dets=None, prefix=None,
                   samples=None, no_signal=None,
                   **kwargs):
        # For Context.get_obs
        if samples is None:
            samples = 0, None
        samples = list(samples)
        if samples[1] is None:
            samples[1] = self.sample_count
        samples = [s if s >= 0 else self.sample_count + s
                   for s in samples]

        tod = core.AxisManager(core.LabelAxis('dets', dets),
                               core.OffsetAxis('samps', samples[1] - samples[0], samples[0]))
        tod.wrap_new('signal', ('dets', 'samps'))
        return tod


if __name__ == '__main__':
    unittest.main()
